# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

import hashlib
from typing import Optional

from xcube.server.api import ApiError, ApiHandler

from .api import api
from .context import S3Context
from .listbucket import (
    _LAST_MODIFIED_DUMMY,
    dict_to_xml,
    list_bucket_result_to_xml,
    list_s3_bucket_v1,
    list_s3_bucket_v2,
)

BUCKET_PARAMETER = {
    "name": "bucket",
    "in": "path",
    "description": "The bucket name."
    ' The bucket "datasets" provides Zarr datasets,'
    ' the bucket "pyramids" provides them as'
    " multi-resolution image pyramids.",
    "schema": {
        "type": "string",
        "enum": ["datasets", "pyramids"],
    },
}

OPTIONAL_STRING_SCHEMA = {
    "type": ["string", "null"],
    "default": None,
}

LIST_BUCKET_PARAMETERS = [
    BUCKET_PARAMETER,
    {
        "name": "list-type",
        "in": "query",
        "description": "The operation version (1 or 2)",
        "schema": {
            "type": ["integer", "null"],
            "enum": [None, 1, 2],
            "default": None,
        },
    },
    {
        "name": "prefix",
        "in": "query",
        "description": "Key prefix",
        "schema": OPTIONAL_STRING_SCHEMA,
    },
    {
        "name": "delimiter",
        "in": "query",
        "description": "Key delimiter",
        "schema": OPTIONAL_STRING_SCHEMA,
    },
    {
        "name": "max-keys",
        "in": "query",
        "description": "Maximum number of keys to return",
        "schema": {
            "type": "integer",
            "default": 1000,
        },
    },
    {
        "name": "start-after",
        "in": "query",
        "description": "Start after given key. For list-type=2 only.",
        "schema": OPTIONAL_STRING_SCHEMA,
    },
    {
        "name": "continuation-token",
        "in": "query",
        "description": "Continuation token for paging. For list-type=2 only.",
        "schema": OPTIONAL_STRING_SCHEMA,
    },
    {
        "name": "marker",
        "in": "query",
        "description": "Continuation token for paging. For list-type=1 only.",
        "schema": OPTIONAL_STRING_SCHEMA,
    },
]

GET_OBJECT_PARAMETERS = [
    BUCKET_PARAMETER,
    {
        "name": "key",
        "in": "path",
        "description": "The object's key.",
        "schema": {
            "type": "string",
        },
    },
]


@api.route("/s3/{bucket}")
class ListS3BucketHandler(ApiHandler[S3Context]):
    @api.operation(
        operationId="listS3Bucket",
        summary="List bucket contents.",
        parameters=LIST_BUCKET_PARAMETERS,
    )
    async def get(self, bucket: str):
        try:
            object_storage = self.ctx.get_bucket(bucket)
        except KeyError:
            return await self._bucket_not_found(bucket)

        list_s3_bucket_params = dict(
            prefix=self.request.get_query_arg("prefix", default=None),
            delimiter=self.request.get_query_arg("delimiter", default=None),
            max_keys=int(self.request.get_query_arg("max-keys", default="1000")),
        )

        list_type = self.request.get_query_arg("list-type", default=None)
        if list_type == "2":  # Most frequently used
            list_s3_bucket = list_s3_bucket_v2
            list_s3_bucket_params.update(
                start_after=self.request.get_query_arg("start-after", default=None),
                continuation_token=self.request.get_query_arg(
                    "continuation-token", default=None
                ),
            )
        elif list_type in (None, "1"):  # Old clients
            list_s3_bucket = list_s3_bucket_v1
            list_s3_bucket_params.update(
                marker=self.request.get_query_arg("marker", default=None)
            )
        else:
            raise ApiError.BadRequest(f"Unknown bucket list type {list_type!r}")

        list_bucket_result = list_s3_bucket(
            object_storage, name=bucket, **list_s3_bucket_params
        )

        xml = list_bucket_result_to_xml(list_bucket_result)
        self.response.set_header("Content-Type", "application/xml")
        await self.response.finish(xml)


@api.route("/s3/{bucket}/{*key}")
class GetS3BucketObjectHandler(ApiHandler[S3Context]):
    @api.operation(
        operationId="getS3ObjectMetadata",
        summary="Get object metadata for given key.",
        parameters=GET_OBJECT_PARAMETERS,
    )
    async def head(self, bucket: str, key: Optional[str]):
        try:
            object_storage = self.ctx.get_bucket(bucket)
        except KeyError:
            return await self._bucket_not_found(bucket)

        try:
            value = object_storage[key]
        except KeyError:
            return await self._key_not_found(key)

        e_tag = hashlib.md5(value).hexdigest()
        content_length = len(value)

        self.response.set_header("ETag", f'"{e_tag}"')
        self.response.set_header("Last-Modified", _LAST_MODIFIED_DUMMY)
        self.response.set_header("Content-Length", str(content_length))
        await self.response.finish()

    # noinspection PyPep8Naming
    @api.operation(
        operationId="getS3ObjectData",
        summary="Get object for given key.",
        parameters=GET_OBJECT_PARAMETERS,
    )
    async def get(self, bucket: str, key: Optional[str]):
        try:
            object_storage = self.ctx.get_bucket(bucket)
        except KeyError:
            return await self._bucket_not_found(bucket)

        try:
            value = object_storage[key]
        except KeyError:
            return await self._key_not_found(key)

        e_tag = hashlib.md5(value).hexdigest()
        content_length = len(value)

        self.response.set_header("ETag", f'"{e_tag}"')
        self.response.set_header("Last-Modified", _LAST_MODIFIED_DUMMY)
        self.response.set_header("Content-Length", str(content_length))
        self.response.set_header("Content-Type", "binary/octet-stream")
        await self.response.finish(value)

    def _key_not_found(self, key: str):
        return self._not_found(
            "NoSuchKey", "The specified key does not exist.", Key=key
        )

    def _bucket_not_found(self, bucket_name: str):
        return self._not_found(
            "NoSuchBucket",
            "The specified bucket does not exist.",
            BucketName=bucket_name,
        )

    def _not_found(self, code: str, message: str, **kwargs):
        self.response.set_header("Content-Type", "application/xml")
        self.response.set_status(404)
        return self.response.finish(
            dict_to_xml(
                root_element_name="Error",
                content_dict={"Code": code, "Message": message, **kwargs},
            )
        )



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

import datetime
from typing import Any, Dict

import xarray as xr

from xcube.constants import FORMAT_NAME_NETCDF4, FORMAT_NAME_ZARR
from xcube.core.gridmapping import GridMapping
from xcube.util.config import NameDictPairList

_TIME_ATTRS_DATA = (
    "time",
    "time_bnds",
    ("time_coverage_start", "time_coverage_end"),
    str,
)


def update_dataset_attrs(
    dataset: xr.Dataset,
    global_attrs: dict[str, Any] = None,
    update_existing: bool = False,
    in_place: bool = False,
) -> xr.Dataset:
    """Update spatio-temporal CF/THREDDS attributes given
    *dataset* according to spatio-temporal coordinate variables
    time, lat, and lon.

    Args:
        dataset: The dataset.
        global_attrs: Optional global attributes.
        update_existing: If ``True``, any existing attributes
            will be updated.
        in_place: If ``True``, *dataset* will be modified in
            place and returned.
    Returns:
        A new dataset, if *in_place* if ``False`` (default),
        else the passed and modified *dataset*.
    """
    if not in_place:
        dataset = dataset.copy()

    if global_attrs:
        dataset.attrs.update(global_attrs)

    dataset = update_dataset_spatial_attrs(
        dataset, update_existing=update_existing, in_place=in_place
    )
    dataset = update_dataset_temporal_attrs(
        dataset, update_existing=update_existing, in_place=in_place
    )
    return dataset


def update_dataset_spatial_attrs(
    dataset: xr.Dataset, update_existing: bool = False, in_place: bool = False
) -> xr.Dataset:
    """Update spatial CF/THREDDS attributes of given *dataset*.

    Args:
        dataset: The dataset.
        update_existing: If ``True``, any existing attributes will be
            updated.
        in_place: If ``True``, *dataset* will be modified in place and
            returned.

    Returns:
        A new dataset, if *in_place* if ``False`` (default), else the
        passed and modified *dataset*.
    """
    if not in_place:
        dataset = dataset.copy()
    gm = GridMapping.from_dataset(dataset)
    gs_attrs = {
        "geospatial_lon_min",
        "geospatial_lon_max",
        "geospatial_lat_min",
        "geospatial_lat_max",
    }
    if update_existing or not gs_attrs.issubset(dataset.attrs):
        # Update dataset with newly retrieved attributes
        dataset.attrs.update(gm.to_dataset_attrs())
        dataset.attrs["date_modified"] = datetime.datetime.now().isoformat()
    return dataset


def update_dataset_temporal_attrs(
    dataset: xr.Dataset, update_existing: bool = False, in_place: bool = False
) -> xr.Dataset:
    """Update temporal CF/THREDDS attributes of given *dataset*.

    Args:
        dataset: The dataset.
        update_existing: If ``True``, any existing attributes will be
            updated.
        in_place: If ``True``, *dataset* will be modified in place and
            returned.

    Returns:
        A new dataset, if *in_place* is ``False`` (default), else the
        passed and modified *dataset*.
    """
    coord_data = [_TIME_ATTRS_DATA]
    if not in_place:
        dataset = dataset.copy()

    for coord_name, coord_bnds_name, coord_attr_names, cast in coord_data:
        coord_min_attr_name, coord_max_attr_name = coord_attr_names
        if (
            update_existing
            or coord_min_attr_name not in dataset.attrs
            or coord_max_attr_name not in dataset.attrs
        ):
            coord = None
            coord_bnds = None
            coord_res = None
            if coord_name in dataset:
                coord = dataset[coord_name]
                coord_bnds_name = coord.attrs.get("bounds", coord_bnds_name)
            if coord_bnds_name in dataset:
                coord_bnds = dataset[coord_bnds_name]
            if (
                coord_bnds is not None
                and coord_bnds.ndim == 2
                and coord_bnds.shape[0] > 0
                and coord_bnds.shape[1] == 2
            ):
                coord_v1 = coord_bnds[0][0]
                coord_v2 = coord_bnds[-1][1]
                coord_res = (coord_v2 - coord_v1) / coord_bnds.shape[0]
                coord_res = float(coord_res.values)
                coord_min, coord_max = (
                    (coord_v1, coord_v2) if coord_res > 0 else (coord_v2, coord_v1)
                )
                dataset.attrs[coord_min_attr_name] = cast(coord_min.values)
                dataset.attrs[coord_max_attr_name] = cast(coord_max.values)
            elif coord is not None and coord.ndim == 1 and coord.shape[0] > 0:
                coord_v1 = coord[0]
                coord_v2 = coord[-1]
                if coord.shape[0] > 1:
                    coord_res = (coord_v2 - coord_v1) / (coord.shape[0] - 1)
                    coord_v1 -= coord_res / 2
                    coord_v2 += coord_res / 2
                    coord_res = float(coord_res.values)
                    coord_min, coord_max = (
                        (coord_v1, coord_v2) if coord_res > 0 else (coord_v2, coord_v1)
                    )
                else:
                    coord_min, coord_max = coord_v1, coord_v2
                dataset.attrs[coord_min_attr_name] = cast(coord_min.values)
                dataset.attrs[coord_max_attr_name] = cast(coord_max.values)

    dataset.attrs["date_modified"] = datetime.datetime.now().isoformat()
    return dataset


def update_dataset_var_attrs(
    dataset: xr.Dataset, var_attrs_list: NameDictPairList
) -> xr.Dataset:
    """Update the attributes of variables in given *dataset*.
    Optionally rename variables according to a given attribute named "name".

    *var_attrs_list* must be a sequence of pairs of the form (<var_name>, <var_attrs>) where <var_name> is a string
    and <var_attrs> is a dictionary representing the attributes to be updated , including an optional "name" attribute.
    If <var_attrs> contains an attribute "name", the variable named <var_name> will be renamed to that attribute's
    value.

    Args:
        dataset: A dataset.
        var_attrs_list: List of tuples of the form (variable name,
            properties dictionary).

    Returns:
        A shallow copy of *dataset* with updated / renamed variables.
    """
    if not var_attrs_list:
        return dataset

    var_name_attrs = dict()
    var_renamings = dict()
    new_var_names = set()

    # noinspection PyUnusedLocal,PyShadowingNames
    for var_name, var_attrs in var_attrs_list:
        if not var_attrs:
            continue
        # noinspection PyShadowingNames
        var_attrs = dict(var_attrs)
        if "name" in var_attrs:
            new_var_name = var_attrs.pop("name")
            if new_var_name in new_var_names:
                raise ValueError(
                    f"variable {var_name!r} cannot be renamed into {new_var_name!r} "
                    "because the name is already in use"
                )
            new_var_names.add(new_var_name)
            var_attrs["original_name"] = var_name
            var_renamings[var_name] = new_var_name
            var_name = new_var_name
        var_name_attrs[var_name] = var_attrs

    if var_renamings:
        dataset = dataset.rename(var_renamings)
    elif var_name_attrs:
        dataset = dataset.copy()

    if var_name_attrs:
        for var_name, var_attrs in var_name_attrs.items():
            var = dataset[var_name]
            var.attrs.update(var_attrs)

    return dataset


def update_dataset_chunk_encoding(
    dataset: xr.Dataset,
    chunk_sizes: dict[str, int] = None,
    format_name: str = None,
    in_place: bool = False,
    data_vars_only: bool = False,
) -> xr.Dataset:
    """Update each variable's encoding in *dataset* with respect to *chunk_sizes*
    so *dataset* is written in chunks for given *format_name*.

    Args:
        dataset: input dataset.
        chunk_sizes: the chunk sizes to be used for the encoding. If
            None, any chunking encoding is removed.
        format_name: format name, e.g. "zarr" or "netcdf4".
        in_place: If ``True``, *dataset* will be modified in place and
            returned.
        data_vars_only: only chunk data variables, not coordinates
    """
    if format_name == FORMAT_NAME_ZARR:
        chunk_sizes_attr_name = "chunks"
    elif format_name == FORMAT_NAME_NETCDF4:
        chunk_sizes_attr_name = "chunksizes"
    else:
        return dataset
    if not in_place:
        dataset = dataset.copy()
    for var_name in dataset.data_vars if data_vars_only else dataset.variables:
        var = dataset[var_name]
        if chunk_sizes is not None:

            def get_size(i):
                dim_name = var.dims[i]
                size = chunk_sizes.get(dim_name)
                if isinstance(size, int):
                    return size
                if var.chunks:
                    size = var.chunks[i]
                    if isinstance(size, int):
                        return size
                    if len(size):
                        return size[0]
                return var.shape[i]

            var.encoding.update(
                {chunk_sizes_attr_name: tuple(map(get_size, range(var.ndim)))}
            )
        elif chunk_sizes_attr_name in var.encoding:
            # Remove any explicit and possibly unintended specification
            del var.encoding[chunk_sizes_attr_name]
    return dataset



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

from abc import abstractmethod
from collections.abc import Sequence
from typing import Any, Dict, Generic, List, Optional, Type, TypeVar

from xcube.core.store import DatasetDescriptor
from xcube.util.assertions import assert_in, assert_instance
from xcube.util.jsonschema import (
    JsonArraySchema,
    JsonIntegerSchema,
    JsonObject,
    JsonObjectSchema,
    JsonStringSchema,
)

STATUS_IDS = ["ok", "error", "warning"]

R = TypeVar("R", bound=JsonObject)


class GenericCubeGeneratorResult(Generic[R], JsonObject):
    def __init__(
        self,
        status: str,
        status_code: Optional[int] = None,
        result: Optional[R] = None,
        message: Optional[str] = None,
        output: Optional[Sequence[str]] = None,
        traceback: Optional[Sequence[str]] = None,
        versions: Optional[dict[str, str]] = None,
    ):
        assert_instance(status, str, name="status")
        assert_in(status, STATUS_IDS, name="status")
        self.status = status
        self.status_code = status_code
        self.result = result
        self.message = message if message else None
        self.output = list(output) if output else None
        self.traceback = list(traceback) if traceback else None
        self.versions = dict(versions) if versions else None

    def derive(
        self,
        /,
        status: Optional[str] = None,
        status_code: Optional[int] = None,
        result: Optional[R] = None,
        message: Optional[str] = None,
        output: Optional[Sequence[str]] = None,
        traceback: Optional[Sequence[str]] = None,
        versions: Optional[dict[str, str]] = None,
    ) -> R:
        return self.__class__(
            status or self.status,
            status_code=status_code or self.status_code,
            result=result or self.result,
            message=message or self.message,
            output=output or self.output,
            traceback=traceback or self.traceback,
            versions=versions or self.versions,
        )

    @classmethod
    @abstractmethod
    def get_result_schema(cls) -> JsonObjectSchema:
        """Get the JSON schema of the result object"""

    @classmethod
    def from_dict(cls, value: dict) -> R:
        return cls.get_schema().from_instance(value)

    @classmethod
    def get_schema(cls) -> JsonObjectSchema:
        return JsonObjectSchema(
            properties=dict(
                status=JsonStringSchema(enum=STATUS_IDS),
                status_code=JsonIntegerSchema(),
                result=cls.get_result_schema(),
                message=JsonStringSchema(),
                output=JsonArraySchema(items=JsonStringSchema()),
                traceback=JsonArraySchema(items=JsonStringSchema()),
                versions=JsonObjectSchema(additional_properties=True),
            ),
            required=["status"],
            additional_properties=True,
            factory=cls,
        )


def make_cube_generator_result_class(
    result_type: type[R],
) -> type[GenericCubeGeneratorResult[R]]:
    class SpecificCubeGeneratorResult(GenericCubeGeneratorResult[R]):
        @classmethod
        def get_result_schema(cls) -> JsonObjectSchema:
            return result_type.get_schema()

    return SpecificCubeGeneratorResult


class CubeReference(JsonObject):
    def __init__(self, data_id: str):
        self.data_id = data_id

    @classmethod
    def get_schema(cls) -> JsonObjectSchema:
        return JsonObjectSchema(
            properties=dict(
                data_id=JsonStringSchema(min_length=1),
            ),
            required=["data_id"],
            additional_properties=False,
            factory=cls,
        )


CubeGeneratorResult = make_cube_generator_result_class(CubeReference)


class CubeInfo(JsonObject):
    def __init__(
        self,
        dataset_descriptor: DatasetDescriptor,
        size_estimation: dict[str, Any],
        **kwargs,
    ):
        self.dataset_descriptor: DatasetDescriptor = dataset_descriptor
        self.size_estimation: dict = size_estimation

    @classmethod
    def get_schema(cls) -> JsonObjectSchema:
        return JsonObjectSchema(
            properties=dict(
                dataset_descriptor=DatasetDescriptor.get_schema(),
                size_estimation=JsonObjectSchema(additional_properties=True),
            ),
            required=["dataset_descriptor", "size_estimation"],
            additional_properties=True,
            factory=cls,
        )


CubeInfoResult = make_cube_generator_result_class(CubeInfo)



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

import os
import warnings
from collections.abc import Container, Iterator
from typing import Any, Union

import fsspec
import xarray as xr

from xcube.util.jsonschema import JsonObjectSchema

from ..datatype import DataType, DataTypeLike
from ..descriptor import DataDescriptor, DatasetDescriptor, new_data_descriptor
from ..store import DataStore
from .schema import REF_STORE_SCHEMA


class ReferenceDataStore(DataStore):
    """A data store that uses kerchunk reference files to
    open referred NetCDF datasets as if they were Zarr datasets.

    Args:
        refs: The list of reference JSON files to use for this instance.
            Files (URLs or local paths) are used in conjunction with
            target_options and target_protocol to open and parse JSON at
            this location.
        target_protocol: Used for loading the reference files. If not
            given, protocol will be derived from the given path.
        target_options: Extra filesystem options for loading the
            reference files.
        remote_protocol: The protocol of the filesystem on which the
            references will be evaluated. If not given, will be derived
            from the first URL in the references that has a protocol.
        remote_options: Extra filesystem options for loading the
            referenced data.
        max_gap: See ``max_block``.
        max_block: For merging multiple concurrent requests to the same
            remote file. Neighboring byte ranges will only be merged
            when their inter-range gap is <= `max_gap`. Default is 64KB.
            Set to 0 to only merge when it requires no extra bytes. Pass
            a negative number to disable merging, appropriate for local
            target files. Neighboring byte ranges will only be merged
            when the size of the aggregated range is <= ``max_block``.
            Default is 256MB.
        cache_size: Maximum size of LRU cache, where
            cache_size*record_size denotes the total number of
            references that can be loaded in memory at once. Only used
            for lazily loaded references.
    """

    def __init__(
        self,
        refs: list[str],
        **ref_kwargs,
    ):
        self._refs = dict(self._normalize_ref(ref) for ref in refs)
        self._ref_kwargs = ref_kwargs

    @classmethod
    def get_data_store_params_schema(cls) -> JsonObjectSchema:
        return REF_STORE_SCHEMA

    @classmethod
    def get_data_types(cls) -> tuple[str, ...]:
        return ("dataset",)

    def get_data_types_for_data(self, data_id: str) -> tuple[str, ...]:
        return self.get_data_types()

    def get_data_ids(
        self,
        data_type: DataTypeLike = None,
        include_attrs: Container[str] | bool = False,
    ) -> Union[Iterator[str], Iterator[tuple[str, dict[str, Any]]]]:
        return iter(self._refs.keys())

    def has_data(self, data_id: str, data_type: DataTypeLike = None) -> bool:
        return data_id in self._refs

    def describe_data(
        self, data_id: str, data_type: DataTypeLike = None
    ) -> DataDescriptor:
        data_descriptor = self._refs[data_id].get("data_descriptor")
        if data_descriptor is None:
            dataset = self.open_data(data_id)
            data_descriptor = new_data_descriptor(data_id, dataset)
            self._refs[data_id]["data_descriptor"] = data_descriptor
        return data_descriptor

    def get_data_opener_ids(
        self, data_id: str = None, data_type: DataTypeLike = None
    ) -> tuple[str, ...]:
        return ("dataset:zarr:reference",)

    def get_open_data_params_schema(
        self, data_id: str = None, opener_id: str = None
    ) -> JsonObjectSchema:
        # We do not have open parameters yet
        return JsonObjectSchema()

    def open_data(
        self, data_id: str, opener_id: str = None, **open_params
    ) -> xr.Dataset:
        data_type = open_params.pop("data_type", None)
        if DataType.normalize(data_type).alias == "mldataset":
            warnings.warn(
                "ReferenceDataStore can only represent the data resource as xr.Dataset."
            )
        if open_params:
            warnings.warn(
                f"open_params are not supported yet,"
                f" but passing forward {', '.join(open_params.keys())}"
            )
        ref_path = self._refs[data_id]["ref_path"]
        open_params.pop("consolidated", False)
        ref_mapping = fsspec.get_mapper("reference://", fo=ref_path, **self._ref_kwargs)
        return xr.open_zarr(ref_mapping, consolidated=False, **open_params)

    @classmethod
    def get_search_params_schema(
        cls, data_type: DataTypeLike = None
    ) -> JsonObjectSchema:
        # We do not have search parameters yet
        return JsonObjectSchema()

    def search_data(
        self, data_type: DataTypeLike = None, **search_params
    ) -> Iterator[DataDescriptor]:
        if search_params:
            warnings.warn(
                f"search_params are not supported yet,"
                f" but received {', '.join(search_params.keys())}"
            )
        return (self.describe_data(data_id) for data_id in self.get_data_ids())

    @classmethod
    def _normalize_ref(
        cls, ref: Union[str, dict[str, Any]]
    ) -> tuple[str, dict[str, Any]]:
        if isinstance(ref, str):
            ref_path = ref
            data_id = cls._ref_path_to_data_id(ref_path)
            return data_id, dict(ref_path=ref_path, data_descriptor=None)

        if isinstance(ref, dict):
            ref_path = ref.get("ref_path") or None
            if not ref_path:
                raise ValueError("missing key ref_path in refs item")

            data_descriptor = ref.get("data_descriptor") or None
            if isinstance(data_descriptor, dict):
                data_descriptor = DatasetDescriptor.from_dict(data_descriptor)
            elif data_descriptor is not None:
                raise TypeError(
                    "value of data_descriptor key in refs item must be a dict or None"
                )

            data_id = ref.get("data_id") or None
            if data_id is None and data_descriptor is not None:
                data_id = data_descriptor.data_id
            if data_id is None:
                data_id = cls._ref_path_to_data_id(ref_path)

            return data_id, dict(ref_path=ref_path, data_descriptor=data_descriptor)
        raise TypeError("item in refs must be a str or a dict")

    @classmethod
    def _ref_path_to_data_id(cls, ref_path: str) -> str:
        protocol, path = fsspec.core.split_protocol(ref_path)
        if protocol in (None, "file", "local") and os.path.sep != "/":
            path = path.replace(os.path.sep, "/")
        name = path.rsplit("/", maxsplit=1)[-1]
        return name[:-5] if name.endswith(".json") else name



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

from .routes import (
    PATH_PARAM_DATASET_ID,
    PATH_PARAM_VAR_NAME,
    QUERY_PARAM_CMAP,
    QUERY_PARAM_CRS,
    QUERY_PARAM_NORM,
    QUERY_PARAM_VMAX,
    QUERY_PARAM_VMIN,
    api,
)



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

import fnmatch
import os.path
import re
import zipfile
from collections.abc import Collection, Iterator
from typing import Any, Optional, Union

import fsspec
import fsspec.implementations.zip
from fsspec.implementations.local import LocalFileSystem

from xcube.core.byoa.constants import TEMP_FILE_PREFIX
from xcube.util.assertions import assert_given, assert_instance
from xcube.util.jsonschema import (
    JsonArraySchema,
    JsonObject,
    JsonObjectSchema,
    JsonStringSchema,
)
from xcube.util.temp import new_temp_dir, new_temp_file


class _FileSetDetails:
    def __init__(
        self, fs: fsspec.AbstractFileSystem, root: str, local_path: Union[None, str]
    ):
        """Stores file system details about a FileSet instance.
        Internal helper class.

        Args:
            fs: file system
            root: root in file system
            local_path: the local or network path. None, if this is not
                a local path.
        """
        self._fs = fs
        self._root = root
        self._local_path = local_path
        self._mapper: Optional[fsspec.FSMap] = None

    @classmethod
    def new(cls, path: str, storage_params: dict[str, Any] = None) -> "_FileSetDetails":
        try:
            fs, root = fsspec.core.url_to_fs(path, **(storage_params or {}))
        except (ImportError, OSError) as e:
            raise ValueError(f"Illegal file set {path!r}") from e
        local_path = (
            root if "file" in fs.protocol or isinstance(fs, LocalFileSystem) else None
        )
        return _FileSetDetails(fs, root, local_path)

    @property
    def fs(self) -> fsspec.AbstractFileSystem:
        return self._fs

    @property
    def root(self) -> str:
        return self._root

    @property
    def local_path(self) -> Optional[str]:
        return self._local_path

    @property
    def mapper(self) -> fsspec.FSMap:
        if self._mapper is None:
            fs, root = self.fs, self.root
            if self.local_path and zipfile.is_zipfile(self.local_path):
                fs = fsspec.implementations.zip.ZipFileSystem(self.local_path)
                root = ""
            self._mapper = fsspec.FSMap(root, fs)
        return self._mapper


class FileSet(JsonObject):
    """A set of files that can found at some abstract root *path*.
    The *path* may identify local or remote filesystems.
    A filesystem may require specific *parameters*, e.g.
    user credentials.

    This implementation is based on the ``fsspec``
    package and the documentation about the different
    possible file systems and their specific parameters
    can be found at https://filesystem-spec.readthedocs.io/.

    Examples:

        FileSet('eurodatacube/my_processor',
                includes=['*.py'])

        FileSet('eurodatacube/my_processor.zip',
                excludes=['*.zarr'])

        FileSet('s3://eurodatacube/my_processor.zip',
                storage_params=dict(key='...',
                                    secret='...')

        FileSet('github://dcs4cop:xcube@v0.8.1/',
                storage_params=dict(username='...',
                                    token='ghp_...')

    If *includes* wildcard patterns are given, only files
    whose paths match any of the patterns are included.
    If *excludes* wildcard patterns are given, any file
    matching one of the patterns is excluded.
    The *includes*, if any, are applied before the
    *excludes*, if any.

    Args:
        path: Root path of the file set.
        sub_path: An optional sub-path relative to *path*. If given,
            this is where the actual file sets starts. For example, the
            only top-level directory entry in a ZIP archive. In archive
            files from GitHub this would be typically "<gh_repo>-<gh-
            release-name>". Note: wildcard characters are not yet
            allowed.
        includes: Wildcard patterns used to include a file.
        excludes: Wildcard patterns used to exclude a file.
        storage_params: File system specific storage parameters.
    """

    def __init__(
        self,
        path: str,
        sub_path: str = None,
        includes: Collection[str] = None,
        excludes: Collection[str] = None,
        storage_params: dict[str, Any] = None,
    ):
        assert_instance(path, str, "path")
        assert_given(path, "path")
        if sub_path is not None:
            assert_instance(sub_path, str, "sub_path")
        self._path = path
        self._sub_path = sub_path
        self._storage_params = (
            dict(storage_params) if storage_params is not None else None
        )
        self._includes = list(includes) if includes is not None else None
        self._excludes = list(excludes) if excludes is not None else None
        # computed members
        self._include_patterns = _translate_patterns(includes or [])
        self._exclude_patterns = _translate_patterns(excludes or [])
        # cached, computed members
        self._details: Optional[_FileSetDetails] = None

    @classmethod
    def get_schema(cls) -> JsonObjectSchema:
        """Get the JSON-schema for FileSet objects."""
        return JsonObjectSchema(
            properties=dict(
                path=JsonStringSchema(min_length=1),
                sub_path=JsonStringSchema(min_length=1),
                storage_params=JsonObjectSchema(additional_properties=True),
                includes=JsonArraySchema(items=JsonStringSchema(min_length=1)),
                excludes=JsonArraySchema(items=JsonStringSchema(min_length=1)),
            ),
            additional_properties=False,
            required=["path"],
            factory=cls,
        )

    @property
    def path(self) -> str:
        """Get the root path."""
        return self._path

    @property
    def sub_path(self) -> Optional[str]:
        """Get the sub-path relative to *path*."""
        return self._sub_path

    @property
    def storage_params(self) -> Optional[dict[str, Any]]:
        """Get optional parameters for the file system
        :attribute:path is referring to.
        """
        return self._storage_params

    @property
    def includes(self) -> Optional[list[str]]:
        """Wildcard patterns used to include a file."""
        return self._includes

    @property
    def excludes(self) -> Optional[list[str]]:
        """Wildcard patterns used to exclude a file."""
        return self._excludes

    def keys(self) -> Iterator[str]:
        """Get keys in this file set.
        A key is a normalized path relative to the root *path*.
        The forward slash "/" is used as path separator.
        """
        for key in self._get_details().mapper.keys():
            # print('-->', key)
            if self._accepts_key(key):
                yield key

    def get_local_path(self) -> Optional[str]:
        """Get the local path, if this file set is local,
        otherwise return None.
        """
        return self._get_details().local_path

    def is_local(self) -> bool:
        """Test whether this file set refers to a local file or directory.
        The test is made on this path's protocol.
        """
        if self._details is not None:
            return self._details.local_path is not None
        protocol, _ = fsspec.core.split_protocol(self._path)
        return protocol is None or protocol == "file"

    def to_local(self) -> "FileSet":
        """Turn this file set into a locally existing file set."""
        if self.is_local():
            return self

        details = self._get_details()
        fs, root = details.fs, details.root

        url_path = fsspec.core.strip_protocol(self.path)
        temp_file_suffix = ""
        for temp_file_suffix in reversed(url_path.split("/")):
            if temp_file_suffix != "":
                break

        if fs.isdir(root):
            temp_dir = new_temp_dir(prefix=TEMP_FILE_PREFIX, suffix=temp_file_suffix)

            # Note, actually want to use
            #    fs.get(root + "/*", temp_dir, recursive=True)
            # here, but fs.get() behaves unpredictably with
            # fsspec=2023.3.0 and s3fs=2023.3.0.
            # Sometimes it will create a new subdirectory in temp_dir
            # named after last path element of root, sometimes not.
            # Behaviour randomly changes if
            #   - temp_dir exists or not
            #   - root or temp_dir paths end with a slash
            #   - whether root or temp_dir are absolute path or not.
            #
            # See https://github.com/dcs4cop/xcube/issues/828
            #
            # Workaround used here is to manually download the directory.

            def get_files(source: str, target: str):
                source_items = fs.listdir(source, detail=True)
                for source_item in source_items:
                    source_path = source_item["name"]
                    source_type = source_item["type"]
                    source_name = source_path.replace("\\", "/").split("/")[-1]
                    target_path = os.path.join(target, source_name)
                    if source_type == "file":
                        fs.get_file(source_path, target_path)
                    elif source_type == "directory":
                        os.mkdir(target_path)
                        get_files(source_path, target_path)

            get_files(root, temp_dir)

            return FileSet(
                temp_dir,
                sub_path=self.sub_path,
                includes=self.includes,
                excludes=self.excludes,
            )
        else:
            _, temp_file = new_temp_file(
                prefix=TEMP_FILE_PREFIX, suffix=temp_file_suffix
            )
            fs.get_file(root, temp_file)
            return FileSet(
                temp_file,
                sub_path=self.sub_path,
                includes=self.includes,
                excludes=self.excludes,
            )

    def is_local_dir(self) -> bool:
        """Test whether this file set refers to an existing local directory."""
        local_path = self.get_local_path()
        return os.path.isdir(local_path) if local_path is not None else False

    def to_local_dir(self, dir_path: str = None) -> "FileSet":
        """Convert this file set into a file set that refers to a
        directory in the local file system.

        The *dir_path* parameter is used only if this fileset
        does not already refer to an existing local directory.

        Args:
            dir_path: An optional directory path. If not given, a
                temporary directory is created.

        Returns:
            The file set representing the local directory.
        """
        file_set = self.to_local()
        if (
            file_set.is_local_dir()
            and dir_path is None
            and not file_set.includes
            and not file_set.excludes
        ):
            return file_set
        return file_set._write_local_dir(dir_path)

    def is_local_zip(self):
        """Test whether this file set refers to an existing local ZIP archive."""
        local_path = self.get_local_path()
        return zipfile.is_zipfile(local_path) if local_path is not None else False

    def to_local_zip(self, zip_path: str = None) -> "FileSet":
        """Zip this file set and return it as a new local directory file set.
        If this is already a ZIP archive, return this file set.

        Args:
            zip_path: An optional path for the new ZIP archive. If not
                given, a temporary file will be created.

        Returns:
            The file set representing the local ZIP archive.
        """
        file_set = self.to_local()
        if (
            file_set.is_local_zip()
            and zip_path is None
            and not file_set._includes
            and not file_set._excludes
        ):
            return file_set
        return file_set._write_local_zip(zip_path)

    def _write_local_dir(self, dir_path: Optional[str]) -> "FileSet":
        """Write file set into local directory."""
        if dir_path:
            if not os.path.exists(dir_path):
                os.makedirs(dir_path, exist_ok=True)
            else:
                # We should assert an empty existing directory
                pass
        else:
            dir_path = new_temp_dir(prefix=TEMP_FILE_PREFIX)

        mapper = self._get_details().mapper

        sub_path = _normalize_sub_path(self.sub_path)

        for key in self.keys():
            file_path = os.path.join(dir_path, _key_to_path(key, sub_path))
            file_dir_path = os.path.dirname(file_path)
            if not os.path.exists(file_dir_path):
                os.makedirs(file_dir_path, exist_ok=True)
            with open(file_path, "wb") as stream:
                stream.write(mapper[key])

        return FileSet(dir_path)

    def _write_local_zip(self, zip_path: Optional[str]) -> "FileSet":
        """Write file set into local ZIP archive."""
        local_dir_path = self.get_local_path()

        if not zip_path:
            _, zip_path = new_temp_file(prefix=TEMP_FILE_PREFIX, suffix=".zip")

        sub_path = _normalize_sub_path(self.sub_path)

        with zipfile.ZipFile(zip_path, "w") as zip_file:
            for key in self.keys():
                file_path = os.path.join(local_dir_path, key)
                zip_file.write(
                    file_path, arcname=_strip_sub_path_from_key(key, sub_path)
                )

        return FileSet(zip_path, sub_path=self.sub_path)

    def _accepts_key(self, key: str) -> bool:
        if key == "":
            return False
        key = "/" + key.replace(os.path.sep, "/")
        return _key_matches(key, self._include_patterns, True) and not _key_matches(
            key, self._exclude_patterns, False
        )

    def _get_details(self) -> _FileSetDetails:
        if self._details is None:
            self._details = _FileSetDetails.new(self.path, self.storage_params)
        return self._details

    def _get_fs(self) -> fsspec.AbstractFileSystem:
        return self._get_details().fs


def _key_matches(key: str, patterns: list[re.Pattern], empty_value: bool) -> bool:
    if not patterns:
        return empty_value
    for pattern in patterns:
        if pattern.match(key):
            return True
    return False


def _translate_patterns(patterns: Optional[Collection[str]]) -> list[re.Pattern]:
    return [
        re.compile(fnmatch.translate(np))
        for p in patterns
        for np in _normalize_pattern(p)
    ]


def _normalize_sub_path(sub_path: str) -> str:
    if sub_path:
        while sub_path.startswith("/"):
            sub_path = sub_path[1:]
        while sub_path.endswith("/"):
            sub_path = sub_path[:-1]
        if sub_path:
            sub_path += "/"
    return sub_path or ""


def _normalize_pattern(pattern: str) -> list[str]:
    pattern = pattern.replace(os.path.sep, "/")
    start_sep = pattern.startswith("/")
    end_sep = pattern.endswith("/")
    if start_sep and end_sep:
        return [pattern, f"{pattern}*"]
    elif start_sep:
        return [pattern, f"{pattern}/*"]
    elif end_sep:
        return [pattern, f"{pattern}*", f"*/{pattern}", f"*/{pattern}*"]
    else:
        return [pattern, f"*/{pattern}", f"*/{pattern}/*"]


def _strip_sub_path_from_key(key: str, sub_path: str):
    if key.startswith(sub_path):
        return key[len(sub_path) :]
    return key


def _key_to_path(key: str, sub_path: str):
    key_path = key
    if sub_path:
        key_path = _strip_sub_path_from_key(key, sub_path)
    return key_path.replace("/", os.path.sep)



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

from collections.abc import Mapping
from typing import Any, List, Optional

from xcube.constants import LOG
from xcube.server.api import ApiError
from xcube.util.jsonschema import (
    JsonArraySchema,
    JsonBooleanSchema,
    JsonObjectSchema,
    JsonStringSchema,
)

AUTHENTICATION_SCHEMA = JsonObjectSchema(
    properties=dict(
        Authority=JsonStringSchema(format="uri"),
        Domain=JsonStringSchema(description="Deprecated, use Authority"),
        Audience=JsonStringSchema(),
        Algorithms=JsonArraySchema(
            items=JsonStringSchema(default=["RS256"]),
        ),
        IsRequired=JsonBooleanSchema(default=False),
    ),
    required=[
        # Either "Authority" or "Domain" are required
        # "Authority",
        "Audience"
    ],
    additional_properties=False,
)


class AuthConfig:
    @classmethod
    def get_schema(cls):
        return JsonObjectSchema(
            properties=dict(Authentication=AUTHENTICATION_SCHEMA),
            factory=AuthConfig.from_dict,
        )

    def __init__(
        self,
        authority: str,
        audience: str,
        algorithms: list[str],
        is_required: bool = False,
    ):
        self._authority = authority
        self._audience = audience
        self._algorithms = algorithms
        self._is_required = is_required

    @property
    def authority(self) -> str:
        return self._authority

    @property
    def _norm_authority(self) -> str:
        return self.authority[:-1] if self.authority.endswith("/") else self.authority

    @property
    def well_known_oid_config(self) -> str:
        return f"{self._norm_authority}/.well-known/openid-configuration"

    @property
    def well_known_jwks(self) -> str:
        return f"{self._norm_authority}/.well-known/jwks.json"

    @property
    def audience(self) -> str:
        return self._audience

    @property
    def algorithms(self) -> list[str]:
        return self._algorithms

    @property
    def is_required(self) -> bool:
        return self._is_required

    @classmethod
    def from_config(cls, config: Mapping[str, Any]) -> Optional["AuthConfig"]:
        authentication = config.get("Authentication")
        if authentication is None:
            return None
        return cls.from_dict(authentication)

    @classmethod
    def from_dict(cls, config: Mapping[str, Any]) -> Optional["AuthConfig"]:
        authority = config.get("Authority")
        domain = config.get("Domain")
        if domain:
            LOG.warning(
                'Configuration parameter "Domain"'
                ' in section "Authentication"'
                ' has been deprecated. Use "Authority" instead.'
            )
            if not authority:
                authority = f"https://{domain}"
        if not authority:
            raise ApiError.InvalidServerConfig(
                'Missing key "Authority" in section "Authentication"'
            )
        audience = config.get("Audience")
        if not audience:
            raise ApiError.InvalidServerConfig(
                'Missing key "Audience" in section "Authentication"'
            )
        algorithms = config.get("Algorithms", ["RS256"])
        if not algorithms:
            raise ApiError.InvalidServerConfig(
                'Value for key "Algorithms" in section'
                ' "Authentication" must not be empty'
            )
        is_required = config.get("IsRequired", False)
        return AuthConfig(authority, audience, algorithms, is_required=is_required)



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

import base64
import io
import json
import os
from abc import ABC, abstractmethod
from collections.abc import Sequence
from functools import cached_property
from typing import Any, Optional, Union

import fsspec
import matplotlib
import matplotlib.colors
import numpy as np
from PIL import Image

from xcube.constants import LOG
from xcube.util.assertions import assert_given, assert_in, assert_instance

try:
    # noinspection PyPackageRequirements
    import cmocean.cm as cmo
except ImportError:
    cmo = None

OCEAN_PREFIX = "cmo"
ALPHA_SUFFIX = "_alpha"
REVERSED_SUFFIX = "_r"  # Used by ocm / Ocean

DEFAULT_CMAP_NAME = "viridis"


# Have colormaps separated into categories taken from
# https://matplotlib.org/stable/gallery/color/colormap_reference.html
# colormaps for ocean taken from
# https://matplotlib.org/cmocean/


class ColormapCategory:
    def __init__(self, name: str, desc: str, cm_names: Optional[list[str]] = None):
        self.name = name
        self.desc = desc
        self.cm_names = cm_names


PERCEPTUALLY_UNIFORM_SEQUENTIAL_CATEGORY = ColormapCategory(
    "Perceptually Uniform Sequential",
    "For many applications, a perceptually uniform colormap"
    " is the best choice - one in which equal steps in data are"
    " perceived as equal steps in the color space",
    ["viridis", "plasma", "inferno", "magma", "cividis"],
)

SEQUENTIAL_CATEGORY = ColormapCategory(
    "Sequential",
    "These colormaps are approximately monochromatic colormaps varying"
    " smoothly between two color tones - usually from low saturation"
    " (e.g. white) to high saturation (e.g. a bright blue)."
    " Sequential colormaps are ideal for representing most scientific"
    " data since they show a clear progression from"
    " low-to-high values.",
    [
        "Greys",
        "Purples",
        "Blues",
        "Greens",
        "Oranges",
        "Reds",
        "YlOrBr",
        "YlOrRd",
        "OrRd",
        "PuRd",
        "RdPu",
        "BuPu",
        "GnBu",
        "PuBu",
        "YlGnBu",
        "PuBuGn",
        "BuGn",
        "YlGn",
    ],
)

SEQUENTIAL_2_CATEGORY = ColormapCategory(
    "Sequential (2)",
    "Many of the values from the Sequential 2 plots are monotonically increasing.",
    [
        "binary",
        "gist_yarg",
        "gist_gray",
        "gray",
        "bone",
        "pink",
        "spring",
        "summer",
        "autumn",
        "winter",
        "cool",
        "Wistia",
        "hot",
        "afmhot",
        "gist_heat",
        "copper",
    ],
)

DIVERGING_CATEGORY = ColormapCategory(
    "Diverging",
    "These colormaps have a median value (usually light in color) and vary"
    " smoothly to two different color tones at high and low values."
    " Diverging colormaps are ideal when your data has a median value"
    " that is significant (e.g.  0, such that positive and negative"
    " values are represented by different colors of the colormap).",
    [
        "PiYG",
        "PRGn",
        "BrBG",
        "PuOr",
        "RdGy",
        "RdBu",
        "RdYlBu",
        "RdYlGn",
        "Spectral",
        "coolwarm",
        "bwr",
        "seismic",
    ],
)

QUALITATIVE_CATEGORY = ColormapCategory(
    "Qualitative",
    "These colormaps vary rapidly in color."
    " Qualitative colormaps are useful for choosing"
    " a set of discrete colors.",
    [
        "Pastel1",
        "Pastel2",
        "Paired",
        "Accent",
        "Dark2",
        "Set1",
        "Set2",
        "Set3",
        "tab10",
        "tab20",
        "tab20b",
        "tab20c",
    ],
)

CYCLIC_CATEGORY = ColormapCategory(
    "Cyclic", "These colormaps are cyclic.", ["twilight", "twilight_shifted", "hsv"]
)

MISCELLANEOUS_CATEGORY = ColormapCategory(
    "Miscellaneous",
    "Colormaps that don't fit into the categories above"
    " or haven't been categorized yet.",
    [
        "flag",
        "prism",
        "ocean",
        "gist_earth",
        "terrain",
        "gist_stern",
        "gnuplot",
        "gnuplot2",
        "CMRmap",
        "cubehelix",
        "brg",
        "gist_rainbow",
        "rainbow",
        "jet",
        "turbo",
        "nipy_spectral",
        "gist_ncar",
    ],
)

OCEAN_CATEGORY = ColormapCategory(
    "Ocean",
    "Colormaps for commonly-used oceanographic variables (from module cmocean.cm).",
)

CUSTOM_CATEGORY = ColormapCategory(
    "Custom",
    (
        "Custom colormaps, e.g. loaded from SNAP *.cpd files or defined via the "
        "xcube server config file."
    ),
)

TEMPLATE_CATEGORIES: dict[str, ColormapCategory] = {
    c.name: c
    for c in (
        PERCEPTUALLY_UNIFORM_SEQUENTIAL_CATEGORY,
        SEQUENTIAL_CATEGORY,
        SEQUENTIAL_2_CATEGORY,
        DIVERGING_CATEGORY,
        QUALITATIVE_CATEGORY,
        CYCLIC_CATEGORY,
        OCEAN_CATEGORY,
        CUSTOM_CATEGORY,
        MISCELLANEOUS_CATEGORY,
    )
}

_CM_NAME_TO_TEMPLATE_CATEGORY = {
    cm_name: template_cat
    for template_cat in TEMPLATE_CATEGORIES.values()
    if template_cat.cm_names
    for cm_name in template_cat.cm_names
}


class Colormap:
    """Internal class (non-API) that represents a color mapping."""

    def __init__(
        self,
        cm_name: str,
        *,
        cm_type: Optional[str] = None,
        cat_name: Optional[str] = None,
        cmap: Optional[matplotlib.colors.Colormap] = None,
        cmap_reversed: Optional[matplotlib.colors.Colormap] = None,
        cmap_alpha: Optional[matplotlib.colors.Colormap] = None,
        norm: Optional[matplotlib.colors.Normalize] = None,
        values: Optional[Sequence[Union[int, float]]] = None,
        cm_code: Optional[dict[str, Any]] = None,
    ):
        assert_instance(cm_name, str)
        if cm_type is None:
            cm_type = "continuous"
        assert_in(cm_type, ("continuous", "stepwise", "categorical"))
        self._cm_name = cm_name
        self._cm_type = cm_type or "continuous"
        self._cat_name = cat_name
        self._cmap = cmap
        self._cmap_reversed = cmap_reversed
        self._cmap_alpha = cmap_alpha
        self._cmap_reversed_alpha = None
        self._cmap_png_base64: Optional[str] = None
        self._norm = norm
        self._colors = values
        self._cm_code = cm_code

    @property
    def cm_name(self) -> str:
        """The colormap's name."""
        return self._cm_name

    @property
    def cm_type(self) -> str:
        """The colormap type, always one of "continuous", "stepwise", "categorical"."""
        return self._cm_type

    @property
    def cat_name(self) -> Optional[str]:
        """The colormap's category name."""
        return self._cat_name

    @property
    def cm_code(self) -> Optional[str]:
        """The colormap's code for generation of custom color maps."""
        return self._cm_code

    @cached_property
    def cmap(self) -> matplotlib.colors.Colormap:
        """The matplotlib colormap."""
        if self._cmap is None:
            self._cmap = matplotlib.colormaps[self.cm_name]
        return self._cmap

    @cached_property
    def cmap_alpha(self) -> matplotlib.colors.Colormap:
        """The matplotlib colormap with alpha-blending of smaller values."""
        if self._cmap_alpha is None:
            _, self._cmap_alpha = get_alpha_cmap(self.cm_name, self.cmap)
        return self._cmap_alpha

    @cached_property
    def cmap_reversed(self) -> matplotlib.colors.Colormap:
        """The reversed matplotlib colormap."""
        """Whether to reverse the color map."""
        if self._cmap_reversed is None:
            _, self._cmap_reversed = get_reverse_cmap(self.cm_name, self.cmap)
        return self._cmap_reversed

    @cached_property
    def cmap_reversed_alpha(self) -> matplotlib.colors.Colormap:
        """The reversed matplotlib colormap with alpha-blending of smaller values."""
        if self._cmap_reversed_alpha is None:
            cm_name, cmap_reversed = get_reverse_cmap(self.cm_name, self.cmap)
            _, self._cmap_reversed_alpha = get_alpha_cmap(cm_name, cmap_reversed)
        return self._cmap_reversed_alpha

    @cached_property
    def cmap_png_base64(self) -> str:
        """The base64 encoding of the colormap's PNG image."""
        if self._cmap_png_base64 is None:
            self._cmap_png_base64 = get_cmap_png_base64(self.cmap)
        return self._cmap_png_base64

    @property
    def norm(self) -> Optional[matplotlib.colors.Normalize]:
        """The matplotlib norm, if any."""
        return self._norm

    @property
    def values(self) -> Optional[Sequence[Union[int, float]]]:
        """The list of values that are mapped to colors."""
        return self._colors


class ColormapProvider(ABC):
    @abstractmethod
    def get_cmap(
        self, cm_name: str, num_colors: Optional[int] = None
    ) -> tuple[matplotlib.colors.Colormap, Colormap]:
        """Get a colormap for the given *cm_name*.

        If *cm_name* is not available, the method may choose another
        colormap and return its name.

        A colormap provider may support colormaps that are

        * reversed (*cm_name* with suffix `"_r"`),
        * have alpha blending (*cm_name* with suffix `"_alpha"`),
        * both (*cm_name* with suffix `"_r_alpha"`).

        Args:
            cm_name: Colormap name.
            num_colors: Optional number of colors, that is, the
                resolution of the colormap gradient.

        Returns:
            A tuple (cmap, colormap) comprising
            the colormap as an instance of
            ``matplotlib.colors.Colormap`` styled according to
            the given *cm_name* suffix(es), and the base colormap object.
        """


class ColormapRegistry(ColormapProvider):
    def __init__(self, *colormaps: Colormap):
        self._categories: dict[str, ColormapCategory] = {}
        self._colormaps: dict[str, Colormap] = {}
        # Add standard Matplotlib colormaps
        self._register_mpl_cmaps()
        # Add Ocean colormaps, if any
        self._register_ocm_cmaps()
        # Add custom colormaps, if any
        for colormap in colormaps:
            self.register_colormap(colormap)

    @property
    def categories(self) -> dict[str, ColormapCategory]:
        return self._categories

    @property
    def colormaps(self) -> dict[str, Colormap]:
        return self._colormaps

    def register_colormap(self, colormap: Colormap):
        cat_name = colormap.cat_name or MISCELLANEOUS_CATEGORY.name
        template_cat = TEMPLATE_CATEGORIES.get(cat_name)
        if template_cat is None:
            raise ValueError(f"Unknown colormap category {cat_name!r}")
        category = self._get_category(template_cat)
        category.cm_names.append(colormap.cm_name)
        self._colormaps[colormap.cm_name] = colormap

    def find_colormap(self, cm_name: str) -> Optional[Colormap]:
        assert_instance(cm_name, str, name="cm_name")
        # Strips suffixes from cm_name
        cm_name, _, _ = parse_cm_name(cm_name)
        return self._colormaps.get(cm_name)

    def get_cmap(
        self, cm_name: str, num_colors: Optional[int] = None
    ) -> tuple[matplotlib.colors.Colormap, Colormap]:
        assert_instance(cm_name, str, name="cm_name")
        if num_colors is not None:
            assert_instance(num_colors, int, name="num_colors")

        colormap: Optional[Colormap] = None
        if cm_name.startswith("{"):
            cm_name, colormap = parse_cm_code(cm_name)

        cm_name, reverse, alpha = parse_cm_name(cm_name)
        if colormap is None:
            colormap = self._colormaps.get(cm_name)

        if colormap is None:
            cm_name = DEFAULT_CMAP_NAME
            colormap = self._colormaps[cm_name]

        if reverse and alpha:
            cmap: matplotlib.colors.Colormap = colormap.cmap_reversed_alpha
        elif reverse:
            cmap: matplotlib.colors.Colormap = colormap.cmap_reversed
        elif alpha:
            cmap: matplotlib.colors.Colormap = colormap.cmap_alpha
        else:
            cmap: matplotlib.colors.Colormap = colormap.cmap
        if num_colors is not None:
            cmap = cmap.resampled(num_colors)
        return cmap, colormap

    def to_json(self) -> list:
        def get_cmap_info(cm_name):
            if self._colormaps[cm_name].cm_code is not None:
                return [
                    cm_name,
                    self._colormaps[cm_name].cmap_png_base64,
                    self._colormaps[cm_name].cm_code,
                ]
            else:
                return [cm_name, self._colormaps[cm_name].cmap_png_base64]

        result = []
        # Loop through TEMPLATE_CATEGORIES to preserve category order
        for cat_name in TEMPLATE_CATEGORIES.keys():
            category = self._categories.get(cat_name)
            if category is not None:
                result.append(
                    [
                        category.name,
                        category.desc,
                        [get_cmap_info(cm_name) for cm_name in category.cm_names],
                    ]
                )
        return result

    def _get_category(self, template_cat: ColormapCategory):
        category = self._categories.get(template_cat.name)
        if category is None:
            category = ColormapCategory(
                template_cat.name, template_cat.desc, cm_names=[]
            )
            self._categories[template_cat.name] = category
        return category

    def _register_mpl_cmaps(self):
        for cm_name, cmap in matplotlib.colormaps.items():
            if cm_name.startswith(OCEAN_PREFIX):
                # We have a separate category for OCEAN maps
                continue
            if cm_name.endswith(REVERSED_SUFFIX):
                # We register reversed maps, if any, separately
                continue
            template_cat = _CM_NAME_TO_TEMPLATE_CATEGORY.get(cm_name)
            if template_cat is None:
                template_cat = MISCELLANEOUS_CATEGORY
            self.register_colormap(
                Colormap(
                    cm_name,
                    cat_name=template_cat.name,
                    cmap=cmap,
                    cmap_reversed=matplotlib.colormaps.get(cm_name + REVERSED_SUFFIX),
                )
            )

    def _register_ocm_cmaps(self):
        if cmo is None:
            return
        for cm_name in cmo.__dict__.keys():
            if not isinstance(
                getattr(cmo, cm_name), matplotlib.colors.Colormap
            ) or cm_name.endswith(REVERSED_SUFFIX):
                continue
            self.register_colormap(
                Colormap(
                    cm_name,
                    cat_name=OCEAN_CATEGORY.name,
                    cmap=getattr(cmo, cm_name),
                    cmap_reversed=getattr(cmo, cm_name + REVERSED_SUFFIX, None),
                )
            )


def parse_cm_code(cm_code: str) -> tuple[str, Optional[Colormap]]:
    # Note, if we get performance issues here, we should
    # cache cm_code -> colormap. Currently, parsing takes less than 50 micros
    try:
        user_color_map: dict[str, Any] = json.loads(cm_code)
        cm_name = user_color_map["name"]
        cm_items = user_color_map["colors"]
        cm_items = [item[:2] for item in cm_items]
        cm_type = user_color_map.get("type", "continuous")
        cm_base_name, _, _ = parse_cm_name(cm_name)
        n = len(cm_items)
        if cm_type == "categorical":
            values: list[int] = []
            colors: list[Union[str, tuple[float, ...]]] = []
            bad = 0, 0, 0, 0
            for i, (value, color) in enumerate(cm_items):
                key = int(value)
                values.append(key)
                colors.append(color)
                if i == n - 1:
                    # the last key's next boundary is key+1
                    values.append(key + 1)
                elif key + 1 != int(cm_items[i + 1][0]):
                    # insert transparent region from key+1 to next key
                    values.append(key + 1)
                    colors.append(bad)
            cmap = matplotlib.colors.ListedColormap(colors, name=cm_base_name)
            cmap.set_extremes(bad=bad, under=bad, over=bad)
        else:  # cm_type == "continuous" or cm_type == "stepwise"
            values, colors = zip(*cm_items)
            vmin = cm_items[0][0]
            vmax = cm_items[-1][0]
            if vmin != 0 or vmax != 1:
                # Normalize values of cm_items between 0 and 1
                norm_values = (np.array(values) - vmin) / (vmax - vmin)
                cm_items = list(zip(norm_values, colors))
            if cm_type == "stepwise":
                # Turn cm_items into discrete step function
                stepwise_cm_items = []
                for i, (value, color) in enumerate(cm_items[0:-1]):
                    next_value = cm_items[i + 1][0]
                    stepwise_cm_items.append((value, color))
                    stepwise_cm_items.append((next_value, color))
                cm_items = stepwise_cm_items
            cmap = matplotlib.colors.LinearSegmentedColormap.from_list(
                cm_base_name, cm_items
            )
        return cm_name, Colormap(
            cm_base_name,
            cm_type=cm_type,
            cat_name=CUSTOM_CATEGORY.name,
            cmap=cmap,
            values=values,
            cm_code=user_color_map,
        )
    except (SyntaxError, KeyError, ValueError, TypeError):
        # If we arrive here, the submitted user-specific cm_code is wrong
        # We do not log or emit a warning here because this would
        # impact performance for current users of xcube-server.
        # Use fallback color map "Reds" to indicate error
        return "Reds", None


def parse_cm_name(cm_name) -> tuple[str, bool, bool]:
    assert_given(cm_name, name="cm_name")
    alpha = cm_name.endswith(ALPHA_SUFFIX)
    if alpha:
        cm_name = cm_name[0 : -len(ALPHA_SUFFIX)]
    reverse = cm_name.endswith(REVERSED_SUFFIX)
    if reverse:
        cm_name = cm_name[0 : -len(REVERSED_SUFFIX)]
    return cm_name, reverse, alpha


def get_reverse_cmap(
    cm_name: str, cmap: matplotlib.colors.Colormap
) -> tuple[str, matplotlib.colors.Colormap]:
    new_cm_name = cm_name + REVERSED_SUFFIX
    return new_cm_name, cmap.reversed(name=new_cm_name)


def get_alpha_cmap(
    cm_name: str, cmap: matplotlib.colors.Colormap
) -> tuple[str, matplotlib.colors.Colormap]:
    """Add extra colormaps with alpha gradient.
    See http://matplotlib.org/api/colors_api.html
    """
    new_cm_name = cm_name + ALPHA_SUFFIX

    # noinspection SpellCheckingInspection
    if isinstance(cmap, matplotlib.colors.LinearSegmentedColormap) and hasattr(
        cmap, "_segmentdata"
    ):
        # noinspection PyProtectedMember
        new_segment_data = dict(cmap._segmentdata)
        # let alpha increase from 0.0 to 0.5
        new_segment_data["alpha"] = ((0.0, 0.0, 0.0), (0.5, 1.0, 1.0), (1.0, 1.0, 1.0))
        return new_cm_name, matplotlib.colors.LinearSegmentedColormap(
            new_cm_name, new_segment_data
        )

    if isinstance(cmap, matplotlib.colors.ListedColormap):
        new_colors = list(cmap.colors)
        a_slope = 2.0 / cmap.N
        a = 0
        for i in range(len(new_colors)):
            # noinspection PyUnresolvedReferences
            new_color = new_colors[i]
            if not isinstance(new_color, str):
                if len(new_color) == 3:
                    r, g, b = new_color
                    new_colors[i] = r, g, b, a
                elif len(new_color) == 4:
                    r, g, b, a_old = new_color
                    new_colors[i] = r, g, b, min(a, a_old)
            a += a_slope
            if a > 1.0:
                a = 1.0
        return new_cm_name, matplotlib.colors.ListedColormap(
            new_colors, name=new_cm_name
        )

    LOG.warning(f"Could not create alpha colormap for cmap of type {type(cmap)}")

    return cm_name, cmap


def get_cmap_png_base64(cmap: matplotlib.colors.Colormap) -> str:
    gradient = np.linspace(0, 1, 256)
    gradient = np.vstack((gradient,))
    image_data = cmap(gradient, bytes=True)
    image = Image.fromarray(image_data, "RGBA")

    # For testing, do:
    # stream = io.FileIO('../cmaps/' + cmap_name + '.png', 'wb')
    # image.save(stream, format='PNG')
    # stream.close()

    stream = io.BytesIO()
    image.save(stream, format="PNG")
    cbar_png_bytes = stream.getvalue()
    stream.close()

    cbar_png_data = base64.b64encode(cbar_png_bytes)
    cbar_png_bytes = cbar_png_data.decode("unicode_escape")

    return cbar_png_bytes


def create_colormap_from_config(cmap_config: dict) -> Colormap:
    registry = ColormapRegistry()
    colors = []
    for color in cmap_config["Colors"]:
        if isinstance(color, list):
            colors.append([c for c in color])
        else:
            if "Label" in color:
                colors.append([color["Value"], color["Color"], color["Label"]])
            else:
                colors.append([color["Value"], color["Color"]])

    for i, item in enumerate(colors):
        if isinstance(item[1], list):
            color = [c for c in item[1]]
            if any([val > 1 for val in color[:3]]):
                color[:3] = np.array(color[:3]) / 255
            colors[i][1] = matplotlib.colors.rgb2hex(color, keep_alpha=True)
    config_parse = dict(
        name=cmap_config["Identifier"],
        type=cmap_config["Type"],
        colors=colors,
    )
    _, cmap = registry.get_cmap(json.dumps(config_parse))
    return cmap, config_parse


def load_custom_colormap(custom_colormap_path: str) -> Optional[Colormap]:
    try:
        # Currently, we only support SNAP *.cpd files
        colormap = load_snap_cpd_colormap(custom_colormap_path)
        LOG.info(f"Loaded custom colormap {custom_colormap_path!r}")
        return colormap
    except FileNotFoundError:
        LOG.error(f"Missing custom colormap file: {custom_colormap_path}")
    except ValueError as e:
        LOG.error(
            f"Detected invalid custom colormap {custom_colormap_path!r}: {e}",
            exc_info=True,
        )
    return None


def load_snap_cpd_colormap(snap_cpd_path: str) -> Colormap:
    points, log_scaled = _parse_snap_cpd_file(snap_cpd_path)
    samples = [sample for sample, _ in points]
    vmin, vmax = min(samples), max(samples)
    if log_scaled:
        norm = matplotlib.colors.LogNorm(vmin, vmax)
    else:
        norm = matplotlib.colors.Normalize(vmin, vmax)
    colors = list(
        zip(map(norm, samples), ["#%02x%02x%02x%02x" % color for _, color in points])
    )
    cm_name, _ = os.path.splitext(os.path.basename(snap_cpd_path))
    return Colormap(
        cm_name,
        cat_name=CUSTOM_CATEGORY.name,
        cmap=matplotlib.colors.LinearSegmentedColormap.from_list(cm_name, colors),
        norm=norm,
    )


Sample = float
Color = tuple[int, int, int, int]
Palette = list[tuple[Sample, Color]]
LogScaled = bool


def _parse_snap_cpd_file(cpd_file_path: str) -> tuple[Palette, LogScaled]:
    with fsspec.open(cpd_file_path, mode="r") as f:
        illegal_format_msg = f"Illegal SNAP *.cpd format: {cpd_file_path}"

        entries: dict[str, str] = dict()
        for line in f.readlines():
            line = line.strip()
            if line and not line.startswith("#"):
                try:
                    k, v = line.split("=", maxsplit=1)
                except ValueError:
                    raise ValueError(illegal_format_msg)
                entries[k.strip()] = v.strip()

        for keyword in ("autoDistribute", "isLogScaled"):
            if keyword in entries:
                LOG.warning(
                    f"Unrecognized keyword {keyword!r}"
                    f" in custom colormap {cpd_file_path}"
                )

        # Uncomment, as soon as we know how to deal with this:
        # log_scaled = entries.get("isLogScaled") == "true"
        log_scaled = False

        try:
            num_points = int(entries.get("numPoints"))
        except ValueError:
            raise ValueError(illegal_format_msg)

        points = []
        for i in range(num_points):
            try:
                rgba = tuple(map(int, entries.get(f"color{i}", "").split(",")))
            except ValueError:
                raise ValueError(illegal_format_msg)
            try:
                sample = float(entries.get(f"sample{i}"))
            except ValueError:
                raise ValueError(illegal_format_msg)
            if len(rgba) == 4:
                points.append((sample, rgba))
            elif len(rgba) == 3:
                points.append((sample, (*rgba, 255)))
            else:
                raise ValueError(illegal_format_msg)

        # noinspection PyTypeChecker
        return points, log_scaled



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

from collections.abc import Sequence
from typing import Any, Optional

import fsspec

from ..assertions import assert_valid_params
from ..error import DataStoreError
from .accessor import FsAccessor, FsDataAccessor
from .impl.dataset import (
    DatasetGeoTiffFsDataAccessor,
    DatasetNetcdfFsDataAccessor,
    DatasetZarrFsDataAccessor,
)
from .impl.fs import (
    AzureFsAccessor,
    FileFsAccessor,
    FtpFsAccessor,
    HttpsFsAccessor,
    MemoryFsAccessor,
    S3FsAccessor,
)
from .impl.geodataframe import (
    GeoDataFrameGeoJsonFsDataAccessor,
    GeoDataFrameShapefileFsDataAccessor,
)
from .impl.geotiff import MultiLevelDatasetGeoTiffFsDataAccessor
from .impl.mldataset import (
    DatasetLevelsFsDataAccessor,
    MultiLevelDatasetLevelsFsDataAccessor,
)
from .store import FsDataStore

############################################
# FsAccessor

_FS_ACCESSOR_CLASSES: dict[str, type[FsAccessor]] = {}


def register_fs_accessor_class(fs_accessor_class: type[FsAccessor]):
    """Register a concrete filesystem accessor class.

    Args:
        fs_accessor_class: a concrete class that extends
            :class:`FsAccessor`.
    """
    protocol = fs_accessor_class.get_protocol()
    _FS_ACCESSOR_CLASSES[protocol] = fs_accessor_class


for cls in (
    AzureFsAccessor,
    FileFsAccessor,
    FtpFsAccessor,
    HttpsFsAccessor,
    MemoryFsAccessor,
    S3FsAccessor,
):
    register_fs_accessor_class(cls)


def get_fs_accessor_class(protocol: str) -> type[FsAccessor]:
    """Get the class for a filesystem accessor.

    Args:
        protocol: The filesystem protocol, for example "file", "s3",
            "memory".

    Returns:
        A class that derives from :class:`FsAccessor`
    """
    fs_accessor_class = _FS_ACCESSOR_CLASSES.get(protocol)
    if fs_accessor_class is None:
        try:
            fsspec.get_filesystem_class(protocol)
        except ImportError as e:
            raise DataStoreError(
                f"Filesystem for protocol {protocol!r}"
                f" is not installed or requires additional packages"
            ) from e
        except ValueError as e:
            raise DataStoreError(
                f"Filesystem not found for protocol {protocol!r}"
            ) from e

        class FsAccessorClass(FsAccessor):
            @classmethod
            def get_protocol(cls) -> str:
                return protocol

        fs_accessor_class = FsAccessorClass
    return fs_accessor_class


############################################
# FsDataAccessor

_FS_DATA_ACCESSOR_CLASSES: dict[str, type[FsDataAccessor]] = {}


def register_fs_data_accessor_class(fs_data_accessor_class: type[FsDataAccessor]):
    """Register an abstract filesystem data accessor class.

    Such data accessor classes are used to dynamically
    construct concrete data store classes by combining
    them with a concrete :class:`FsAccessor`.

    Args:
        fs_data_accessor_class: an abstract class that extends
            :class:`FsDataAccessor`.
    """
    data_type = fs_data_accessor_class.get_data_type()
    format_id = fs_data_accessor_class.get_format_id()
    key = f"{data_type.alias}:{format_id}"
    _FS_DATA_ACCESSOR_CLASSES[key] = fs_data_accessor_class


for cls in (
    DatasetZarrFsDataAccessor,
    DatasetNetcdfFsDataAccessor,
    DatasetGeoTiffFsDataAccessor,
    DatasetLevelsFsDataAccessor,
    MultiLevelDatasetGeoTiffFsDataAccessor,
    MultiLevelDatasetLevelsFsDataAccessor,
    GeoDataFrameShapefileFsDataAccessor,
    GeoDataFrameGeoJsonFsDataAccessor,
):
    register_fs_data_accessor_class(cls)


def get_fs_data_accessor_class(
    protocol: str, data_type_alias: str, format_id: str
) -> type[FsDataAccessor]:
    """Get the class for a filesystem data accessor.

    Args:
        protocol: The filesystem protocol, for example "file", "s3",
            "memory".
        data_type_alias: The data type alias name, for example
            "dataset", "geodataframe".
        format_id: The format identifier, for example "zarr", "geojson".

    Returns:
        A class that derives from :class:`FsAccessor`
    """
    accessor_id = f"{data_type_alias}:{format_id}"
    data_accessor_class = _FS_DATA_ACCESSOR_CLASSES.get(accessor_id)
    if data_accessor_class is None:
        raise DataStoreError(
            f"Combination of data type {data_type_alias!r}"
            f" and format {format_id!r} is not supported"
        )

    fs_accessor_class = get_fs_accessor_class(protocol)

    class FsDataAccessorClass(fs_accessor_class, data_accessor_class):
        pass

    # Should we set __name_ and __doc__ properties here?

    return FsDataAccessorClass


############################################
# FsDataStore


def get_fs_data_store_class(protocol: str) -> type[FsDataStore]:
    """Get the class for of a filesystem-based data store.

    Args:
        protocol: The filesystem protocol, for example "file", "s3",
            "memory".

    Returns:
        A class that derives from :class:`FsDataStore`
    """
    fs_accessor_class = get_fs_accessor_class(protocol)

    class FsDataStoreClass(fs_accessor_class, FsDataStore):
        pass

    # Should we set set __name_ and __doc__ properties here?

    return FsDataStoreClass


def new_fs_data_store(
    protocol: str,
    root: str = "",
    max_depth: Optional[int] = 1,
    read_only: bool = False,
    includes: Optional[Sequence[str]] = None,
    excludes: Optional[Sequence[str]] = None,
    storage_options: dict[str, Any] = None,
) -> FsDataStore:
    """Create a new instance of a filesystem-based data store.

    The data store is capable of filtering the data identifiers reported
    by ``get_data_ids()``. For this purpose the optional keywords
    `excludes` and `includes` are used which can both take the form of
    a wildcard pattern or a sequence of wildcard patterns:

    * ``excludes``: if given and if any pattern matches the identifier,
      the identifier is not reported.
    * ``includes``: if not given or if any pattern matches the identifier,
      the identifier is reported.

    Args:
        protocol: The filesystem protocol, for example "file", "s3",
            "memory".
        root: Root or base directory. Defaults to "".
        max_depth: Maximum recursion depth. None means limitless.
            Defaults to 1.
        read_only: Whether this is a read-only store. Defaults to False.
        includes: Optional sequence of wildcards that include certain
            filesystem paths. Affects the data identifiers (paths)
            returned by `get_data_ids()`. By default, all paths are
            included.
        excludes: Optional sequence of wildcards that exclude certain
            filesystem paths. Affects the data identifiers (paths)
            returned by `get_data_ids()`. By default, no paths are
            excluded.
        storage_options: Options specific to the underlying filesystem
            identified by *protocol*. Used to instantiate the
            filesystem.

    Returns:
        A new data store instance of type :class:`FsDataStore`.
    """
    fs_data_store_class = get_fs_data_store_class(protocol)
    store_params_schema = fs_data_store_class.get_data_store_params_schema()
    store_params = {
        k: v
        for k, v in dict(
            root=root,
            max_depth=max_depth,
            read_only=read_only,
            includes=includes,
            excludes=excludes,
            storage_options=storage_options,
        ).items()
        if v is not None
    }
    assert_valid_params(store_params, name="store_params", schema=store_params_schema)
    return fs_data_store_class(**store_params)



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

import collections.abc
import threading
import warnings
from typing import Optional

import xarray as xr

from xcube.constants import LOG
from xcube.util.assertions import assert_instance


@xr.register_dataset_accessor("zarr_store")
class ZarrStoreHolder:
    """Represents a xarray dataset property ``zarr_store``.

    It is used to permanently associate a dataset with its
    Zarr store, which would otherwise not be possible.

    In xcube server, we use the new property to expose
    datasets via the S3 emulation API.

    For that concept to work, datasets must be associated
    with their Zarr stores explicitly.
    Therefore, the xcube data store framework sets the
    Zarr stores of datasets after opening them ``xr.open_zarr()``:::

        dataset = xr.open_zarr(zarr_store, **open_params)
        dataset.zarr_store.set(zarr_store)

    Note, that the dataset may change after the Zarr store has been set,
    so that the dataset and its Zarr store are no longer in sync.
    This may be an issue and limit the application of the new property.

    Args:
        dataset: The xarray dataset that is associated with a Zarr
            store.
    """

    def __init__(self, dataset: xr.Dataset):
        self._dataset = dataset
        self._zarr_store: Optional[collections.abc.MutableMapping] = None
        self._lock = threading.RLock()

    def get(self) -> collections.abc.MutableMapping:
        """Get the Zarr store of a dataset.
        If no Zarr store has been set, the method will use
        ``GenericZarrStore.from_dataset()`` to create and set
        one.

        Returns:
            The Zarr store.
        """
        if self._zarr_store is None:
            # Double-checked locking pattern
            with self._lock:
                if self._zarr_store is None:
                    from xcube.core.zarrstore import GenericZarrStore

                    self._zarr_store = GenericZarrStore.from_dataset(self._dataset)
                    source = self._dataset.encoding.get("source", "?")
                    LOG.warning(
                        f"dataset {source!r} is assigned a"
                        f" GenericZarrStore which may introduce"
                        f" performance penalties"
                    )
        return self._zarr_store

    def set(self, zarr_store: collections.abc.MutableMapping) -> None:
        """Set the Zarr store of a dataset.

        Args:
            zarr_store: The Zarr store.
        """
        assert_instance(zarr_store, collections.abc.MutableMapping, name="zarr_store")
        with self._lock:
            self._zarr_store = zarr_store

    def reset(self) -> None:
        """Resets the Zarr store."""
        with self._lock:
            self._zarr_store = None



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

import pathlib
from collections.abc import Iterator

import fsspec
from fsspec.implementations.http import HTTPFileSystem
from fsspec.implementations.local import LocalFileSystem


def is_local_fs(fs: fsspec.AbstractFileSystem) -> bool:
    """Check whether *fs* is a local filesystem."""
    return (
        "file" in fs.protocol
        or "local" in fs.protocol
        or isinstance(fs, LocalFileSystem)
    )


def is_https_fs(fs: fsspec.AbstractFileSystem) -> bool:
    """Check whether *fs* is a local filesystem."""
    return "https" in fs.protocol or isinstance(fs, HTTPFileSystem)


def get_fs_path_class(fs: fsspec.AbstractFileSystem) -> type[pathlib.PurePath]:
    """Get the appropriate ``pathlib.PurePath`` class for the filesystem *fs*."""
    if is_local_fs(fs):
        # Will return PurePosixPath or a PureWindowsPath object
        return pathlib.PurePath
    # PurePosixPath for non-local filesystems such as S3,
    # so we force use of forward slashes as separators
    return pathlib.PurePosixPath


def resolve_path(path: pathlib.PurePath) -> pathlib.PurePath:
    """Resolve "." and ".." occurrences in *path* without I/O and
    return a new path.
    """
    reversed_parts = reversed(path.parts)
    reversed_norm_parts = list(_resolve_path_impl(reversed_parts))
    parts = reversed(reversed_norm_parts)
    return type(path)(*parts)


def _resolve_path_impl(reversed_parts: Iterator[str]):
    skips = False
    for part in reversed_parts:
        if part == ".":
            continue
        elif part == "..":
            skips += 1
            continue
        if skips == 0:
            yield part
        elif skips > 0:
            skips -= 1
    if skips != 0:
        raise ValueError('cannot resolve path, misplaced ".."')



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

import abc
from collections.abc import Sequence
from typing import Dict, List, Optional, Union

from xcube.util.assertions import assert_false, assert_given


class Node(abc.ABC):
    """An abstract XML node."""

    @abc.abstractmethod
    def add(self, *elements: "Element") -> "Node":
        """Adds child elements to this element.

        Args:
            *elements: Child elements
        """
        pass

    @abc.abstractmethod
    def to_xml(self, indent: int = 2) -> str:
        """Converts this node into valid XML text using UTF-8 encoding.

        Args:
            indent: Indent in spaces. Defaults to 2.
        """
        pass

    def __str__(self):
        """Calls to_xml() with default arguments."""
        return self.to_xml()


class Element(Node):
    """An XML element node.

    Args:
        tag: Tag name
        attrs: Attributes
        text: Text
        elements: Child elements
    """

    def __init__(
        self,
        tag: str,
        attrs: Optional[dict[str, str]] = None,
        text: Optional[Union[str, Sequence[str]]] = None,
        elements: Optional[Sequence["Element"]] = None,
    ):
        assert_given(tag, name="tag")
        assert_false(
            text and elements, message="text and elements are mutually exclusive"
        )
        self._tag = tag
        self._text = text
        self._attrs = dict(attrs) if attrs else {}
        self._elements = list(elements) if elements else []

    def add(self, *elements: "Element") -> "Element":
        self._elements.extend(elements)
        return self

    def to_xml(self, indent: int = 2) -> str:
        lines = []
        self._to_xml(indent, 0, lines)
        return "\n".join(lines)

    def _to_xml(self, indent: int, level: int, lines: list[str]):
        tab = indent * " "
        tabs = level * tab
        line = f"{tabs}<{self._tag}"
        attrs = self._attrs
        text = self._text
        elements = self._elements
        if attrs:
            for k, v in attrs.items():
                line += f' {k}="{v}"'
        if text:
            if isinstance(text, str):
                lines.append(f"{line}>{text}</{self._tag}>")
            else:
                lines.append(line + ">")
                for t in text:
                    lines.append(f"{tabs}{tab}{t}")
                lines.append(f"{tabs}</{self._tag}>")
        elif elements:
            lines.append(line + ">")
            for node in elements:
                node._to_xml(indent, level + 1, lines)
            lines.append(f"{tabs}</{self._tag}>")
        else:
            lines.append(line + "/>")


class Document(Node):
    """An XML document.

    Args:
        root: The only root element.
    """

    def __init__(self, root: Element):
        self._root = root

    def add(self, *elements: "Element") -> "Document":
        self._root.add(*elements)
        return self

    def to_xml(self, indent: int = 2) -> str:
        xml = self._root.to_xml()
        return f'<?xml version="1.0" encoding="UTF-8"?>\n{xml}\n'



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

import warnings
from collections.abc import Sequence
from typing import AbstractSet, Optional, Union

import numpy as np
import pyproj
import shapely.geometry
import shapely.ops
import shapely.wkt
import xarray as xr

from xcube.constants import CRS_CRS84
from xcube.core.geom import (
    GeometryLike,
    get_dataset_geometry,
    mask_dataset_by_geometry,
    normalize_geometry,
)
from xcube.core.gridmapping import GridMapping
from xcube.core.varexpr import VarExprContext, split_var_assignment
from xcube.util.assertions import assert_instance
from xcube.util.timeindex import ensure_time_index_compatible

Date = Union[np.datetime64, str]

AGG_MEAN = "mean"
AGG_MEDIAN = "median"
AGG_STD = "std"
AGG_MIN = "min"
AGG_MAX = "max"
AGG_COUNT = "count"

MUST_LOAD = True
CAN_COMPUTE = False

AGG_METHODS = {
    AGG_MEAN: CAN_COMPUTE,
    AGG_MEDIAN: MUST_LOAD,
    AGG_STD: CAN_COMPUTE,
    AGG_MIN: CAN_COMPUTE,
    AGG_MAX: CAN_COMPUTE,
    AGG_COUNT: CAN_COMPUTE,
}


def get_time_series(
    dataset: xr.Dataset,
    grid_mapping: Optional[GridMapping] = None,
    geometry: Optional[GeometryLike] = None,
    var_names: Optional[Sequence[str]] = None,
    start_date: Optional[Date] = None,
    end_date: Optional[Date] = None,
    agg_methods: Union[str, Sequence[str], AbstractSet[str]] = AGG_MEAN,
    use_groupby: bool = False,
    cube_asserted: Optional[bool] = None,
) -> Optional[xr.Dataset]:
    """Get a time series dataset from a data *cube*.

    *geometry* may be provided as a (shapely) geometry object, a valid
    GeoJSON object, a valid WKT string,
    a sequence of box coordinates (x1, y1, x2, y2), or point coordinates
    (x, y). If *geometry* covers an area,
    i.e. is not a point, the function aggregates the variables to compute a
    mean value and if desired,
    the number of valid observations and the standard deviation.

    *start_date* and *end_date* may be provided as a numpy.datetime64 or an
    ISO datetime string.

    Returns a time-series dataset whose data variables have a time dimension
    but no longer have spatial dimensions,
    hence the resulting dataset's variables will only have N-2 dimensions.
    A global attribute ``max_number_of_observations`` will be set to the
    maximum number of observations
    that could have been made in each time step.
    If the given *geometry* does not overlap the cube's boundaries, or if not
    output variables remain,
    the function returns ``None``.

    Args:
        dataset: The dataset
        grid_mapping: Grid mapping of *cube*.
        geometry: Optional geometry
        var_names: Optional sequence of names of variables to be
            included.
        start_date: Optional start date.
        end_date: Optional end date.
        agg_methods: Aggregation methods. May be single string or
            sequence of strings. Possible values are 'mean', 'median',
            'min', 'max', 'std', 'count'. Defaults to 'mean'. Ignored if
            geometry is a point.
        use_groupby: Use group-by operation. May increase or decrease
            runtime performance and/or memory consumption.
        cube_asserted: Deprecated and ignored since xcube 0.11.0. No
            replacement.
    """
    if cube_asserted is not None:
        warnings.warn(
            "cube_asserted has been deprecated and will be removed soon.",
            DeprecationWarning,
        )
    assert_instance(dataset, xr.Dataset)
    if grid_mapping is not None:
        assert_instance(grid_mapping, GridMapping)
    else:
        grid_mapping = GridMapping.from_dataset(dataset)

    geometry = normalize_geometry(geometry)
    if geometry is not None and not grid_mapping.crs.is_geographic:
        project = pyproj.Transformer.from_crs(
            CRS_CRS84, grid_mapping.crs, always_xy=True
        ).transform
        geometry = shapely.ops.transform(project, geometry)

    if var_names is not None:
        data_vars: dict[str, xr.DataArray] = {}
        for var_name_or_assign in var_names:
            var_name, var_expr = split_var_assignment(var_name_or_assign)
            if var_expr:
                variable = VarExprContext(dataset).evaluate(var_expr)
            else:
                var_name = var_name_or_assign
                variable = dataset[var_name]

            if isinstance(variable, xr.DataArray) and "time" in variable.dims:
                data_vars[var_name] = variable

        dataset = xr.Dataset(data_vars)

    if len(dataset.data_vars) == 0:
        return None

    if start_date is not None or end_date is not None:
        date_slice = slice(start_date, end_date)
        safe_slice = ensure_time_index_compatible(dataset, date_slice)
        dataset = dataset.sel(time=safe_slice)

    x_name, y_name = grid_mapping.xy_dim_names
    if isinstance(geometry, shapely.geometry.Point):
        bounds = get_dataset_geometry(dataset)
        if not bounds.contains(geometry):
            return None
        indexers = {x_name: geometry.x, y_name: geometry.y}
        dataset = dataset.sel(**indexers, method="Nearest")
        return dataset.assign_attrs(max_number_of_observations=1)

    agg_methods = normalize_agg_methods(agg_methods)

    if geometry is not None:
        dataset = mask_dataset_by_geometry(
            dataset,
            geometry,
            update_attrs=False,
            save_geometry_mask="__mask__",
        )
        if dataset is None:
            return None
        mask = dataset["__mask__"]
        max_number_of_observations = np.count_nonzero(mask)
        dataset = dataset.drop_vars(["__mask__"])
    else:
        max_number_of_observations = dataset[y_name].size * dataset[x_name].size

    must_load = len(agg_methods) > 1 or any(
        AGG_METHODS[agg_method] == MUST_LOAD for agg_method in agg_methods
    )
    if must_load:
        dataset.load()

    agg_datasets = []
    if use_groupby:
        time_group = dataset.groupby("time")
        for agg_method in agg_methods:
            method = getattr(time_group, agg_method)
            if agg_method == "count":
                agg_dataset = method(dim=xr.ALL_DIMS)
            else:
                agg_dataset = method(dim=xr.ALL_DIMS, skipna=True)
            agg_datasets.append(agg_dataset)
    else:
        for agg_method in agg_methods:
            method = getattr(dataset, agg_method)
            if agg_method == "count":
                agg_dataset = method(dim=(y_name, x_name))
            else:
                agg_dataset = method(dim=(y_name, x_name), skipna=True)
            agg_datasets.append(agg_dataset)

    agg_datasets = [
        agg_dataset.rename(
            name_dict={v: f"{v}_{agg_method}" for v in agg_dataset.data_vars}
        )
        for agg_method, agg_dataset in zip(agg_methods, agg_datasets)
    ]

    ts_dataset = xr.merge(agg_datasets)
    ts_dataset = ts_dataset.assign_attrs(
        max_number_of_observations=max_number_of_observations
    )

    return ts_dataset


def normalize_agg_methods(
    agg_methods: Union[str, Sequence[str]], exception_type=ValueError
) -> set[str]:
    agg_methods = agg_methods or [AGG_MEAN]
    if isinstance(agg_methods, str):
        agg_methods = [agg_methods]
    agg_methods = set(agg_methods)
    invalid_agg_methods = agg_methods - set(AGG_METHODS.keys())
    if invalid_agg_methods:
        s = "s" if len(invalid_agg_methods) > 1 else ""
        raise exception_type(
            f"invalid aggregation method{s}:"
            f" {', '.join(sorted(list(invalid_agg_methods)))}"
        )
    return agg_methods



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.
import math

import pyproj
import xarray as xr

from xcube.core.gridmapping import GridMapping
from xcube.core.select import (
    select_spatial_subset,
    select_temporal_subset,
    select_variables_subset,
)

from ..config import CubeConfig
from .transformer import CubeTransformer, TransformedCube


class CubeSubsetter(CubeTransformer):
    def transform_cube(
        self, cube: xr.Dataset, gm: GridMapping, cube_config: CubeConfig
    ) -> TransformedCube:
        desired_var_names = cube_config.variable_names
        if desired_var_names:
            cube = select_variables_subset(cube, var_names=desired_var_names)
            cube_config = cube_config.drop_props("variable_names")

        desired_bbox = cube_config.bbox
        if desired_bbox is not None:
            # Find out whether its possible to make a spatial subset
            # without resampling. First, grid mapping must be regular.
            can_do_spatial_subset = False
            if gm.is_regular:
                can_do_spatial_subset = True
                # Current spatial resolution must be the
                # desired spatial resolution, otherwise spatial resampling
                # is required later, which will include the desired
                # subsetting.
                desired_res = cube_config.spatial_res
                if desired_res is not None and not (
                    math.isclose(gm.x_res, desired_res)
                    and math.isclose(gm.y_res, desired_res)
                ):
                    can_do_spatial_subset = False
                if can_do_spatial_subset:
                    # Finally, the desired CRS must be equal to the current
                    # one, or they must both be geographic.
                    desired_crs = cube_config.crs
                    if desired_crs:
                        desired_crs = pyproj.CRS.from_string(desired_crs)
                        if desired_crs != gm.crs and not (
                            desired_crs.is_geographic and gm.crs.is_geographic
                        ):
                            can_do_spatial_subset = False
            if can_do_spatial_subset:
                cube = select_spatial_subset(cube, xy_bbox=desired_bbox)
                # Now that we have a new cube subset, we must adjust
                # its grid mapping.
                gm = GridMapping.from_dataset(
                    cube,
                    crs=gm.crs,
                )
                # Consume spatial properties
                cube_config = cube_config.drop_props(["bbox", "spatial_res", "crs"])

        desired_time_range = cube_config.time_range
        if desired_time_range:
            cube = select_temporal_subset(cube, time_range=desired_time_range)
            cube_config = cube_config.drop_props("time_range")

        return cube, gm, cube_config



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

import math
from collections.abc import Hashable, Mapping, Sequence
from typing import Any, Callable, Optional, Union

import numpy as np
import xarray as xr
from dask import array as da
from dask_image import ndinterp

from xcube.core.gridmapping import GridMapping
from xcube.core.gridmapping.helpers import AffineTransformMatrix
from xcube.util.assertions import assert_true

from .cf import complete_resampled_dataset

NDImage = Union[np.ndarray, da.Array]
Aggregator = Callable[[NDImage], NDImage]


def affine_transform_dataset(
    source_ds: xr.Dataset,
    /,
    source_gm: Optional[GridMapping] = None,
    target_gm: Optional[GridMapping] = None,
    ref_ds: Optional[xr.Dataset] = None,
    var_configs: Optional[Mapping[Hashable, Mapping[str, Any]]] = None,
    encode_cf: bool = True,
    gm_name: Optional[str] = None,
    reuse_coords: bool = False,
) -> xr.Dataset:
    """Resample dataset according to an affine transformation.

    The affine transformation will be applied only if the CRS of
    *source_gm* and the CRS of *target_gm* are both geographic or equal.
    Otherwise, a ``ValueError`` will be raised.

    New in 1.6: If *target_ds* is given, its coordinate
    variables are copied by reference into the returned
    dataset.

    Args:
        source_ds: The source dataset
        source_gm: Optional source grid mapping of *dataset*.
            If not provided, computed from *source_ds*.
            Must be regular and must have same CRS as *target_gm*.
        target_gm: Optional target grid mapping. If not provided,
            computed from *target_ds* or source grid mapping.
            Must be regular and must have same CRS as source grid mapping.
        ref_ds: An optional dataset that provides the
            target grid mapping if *target_gm* is not provided.
            If *ref_ds* is given, its coordinate variables are copied
            by reference into the returned dataset.
        var_configs: Optional resampling configurations for individual
            variables.
        encode_cf: Whether to encode the target grid mapping into the
            resampled dataset in a CF-compliant way. Defaults to
            ``True``.
        gm_name: Name for the grid mapping variable. Defaults to "crs".
            Used only if *encode_cf* is ``True``.
        reuse_coords: Whether to either reuse target coordinate arrays
            from target_gm or to compute new ones.

    Returns:
        The resampled target dataset.
    """
    if source_gm is None:
        # No source grid mapping given, so do derive it from dataset
        source_gm = GridMapping.from_dataset(source_ds)

    if target_gm is None:
        # No target grid mapping given, so do derive it
        # from reference dataset or source grid mapping
        if ref_ds is not None:
            target_gm = GridMapping.from_dataset(ref_ds)
        else:
            target_gm = source_gm.to_regular()

    # Are source and target both geographic grid mappings?
    both_geographic = source_gm.crs.is_geographic and target_gm.crs.is_geographic
    if not (both_geographic or source_gm.crs == target_gm.crs):
        raise ValueError(
            f"CRS of source_gm and target_gm must be equal,"
            f' was "{source_gm.crs.name}"'
            f' and "{target_gm.crs.name}"'
        )
    GridMapping.assert_regular(source_gm, name="source_gm")
    GridMapping.assert_regular(target_gm, name="target_gm")
    resampled_dataset = resample_dataset(
        dataset=source_ds,
        matrix=target_gm.ij_transform_to(source_gm),
        size=target_gm.size,
        tile_size=target_gm.tile_size,
        xy_dim_names=source_gm.xy_dim_names,
        var_configs=var_configs,
    )
    has_bounds = any(
        source_ds[var_name].attrs.get("bounds") for var_name in source_gm.xy_var_names
    )
    new_coords = target_gm.to_coords(
        xy_var_names=source_gm.xy_var_names,
        xy_dim_names=source_gm.xy_dim_names,
        exclude_bounds=not has_bounds,
        reuse_coords=reuse_coords,
    )
    return complete_resampled_dataset(
        encode_cf,
        resampled_dataset.assign_coords(new_coords),
        target_gm,
        gm_name,
        ref_ds.coords if ref_ds else None,
    )


def resample_dataset(
    dataset: xr.Dataset,
    matrix: AffineTransformMatrix,
    size: tuple[int, int],
    tile_size: tuple[int, int],
    xy_dim_names: tuple[str, str],
    var_configs: Mapping[Hashable, Mapping[str, Any]] = None,
) -> xr.Dataset:
    """Resample dataset according to an affine transformation.

    Args:
        dataset: The source dataset
        matrix: Affine transformation matrix.
        size: Target image size.
        tile_size: Target image tile size.
        xy_dim_names: Names of the spatial dimensions.
        var_configs: Optional resampling configurations for individual
            variables.

    Returns:
        The resampled target dataset.
    """
    ((i_scale, _, i_off), (_, j_scale, j_off)) = matrix
    width, height = size
    tile_width, tile_height = tile_size
    x_dim, y_dim = xy_dim_names
    yx_dims = (y_dim, x_dim)
    coords = dict()
    var_configs = var_configs or {}
    data_vars = dict()
    for k, var in dataset.variables.items():
        new_var = None
        if var.ndim >= 2 and var.dims[-2:] == yx_dims:
            var_config = var_configs.get(k, dict())
            if np.issubdtype(var.dtype, np.integer) or np.issubdtype(var.dtype, bool):
                spline_order = 0
                aggregator = None
                recover_nan = False
            else:
                spline_order = 1
                aggregator = np.nanmean
                # forman: changed default from True to False (v1.6, 2024-06-05)
                recover_nan = False
            var_data = resample_ndimage(
                var.data,
                scale=(j_scale, i_scale),
                offset=(j_off, i_off),
                shape=(height, width),
                chunks=(tile_height, tile_width),
                spline_order=var_config.get("spline_order", spline_order),
                aggregator=var_config.get("aggregator", aggregator),
                recover_nan=var_config.get("recover_nan", recover_nan),
            )
            new_var = xr.DataArray(var_data, dims=var.dims, attrs=var.attrs)
        elif x_dim not in var.dims and y_dim not in var.dims:
            new_var = var.copy()
        if new_var is not None:
            if k in dataset.coords:
                coords[k] = new_var
            elif k in dataset.data_vars:
                data_vars[k] = new_var

    return xr.Dataset(data_vars=data_vars, coords=coords, attrs=dataset.attrs)


def resample_ndimage(
    image: NDImage,
    scale: Union[float, tuple[float, float]] = 1,
    offset: Union[float, tuple[float, float]] = None,
    shape: Union[int, tuple[int, int]] = None,
    chunks: Sequence[int] = None,
    spline_order: int = 1,
    aggregator: Optional[Aggregator] = np.nanmean,
    recover_nan: bool = False,
) -> da.Array:
    image = da.asarray(image)
    offset = _normalize_offset(offset, image.ndim)
    scale = _normalize_scale(scale, image.ndim)
    if shape is None:
        shape = resize_shape(image.shape, scale)
    else:
        shape = _normalize_shape(shape, image)
    chunks = _normalize_chunks(chunks, shape)
    scale_y, scale_x = scale[-2], scale[-1]
    divisor_x = math.ceil(abs(scale_x))
    divisor_y = math.ceil(abs(scale_y))
    if (divisor_x >= 2 or divisor_y >= 2) and aggregator is not None:
        # Downsampling
        # ------------
        axes = {image.ndim - 2: divisor_y, image.ndim - 1: divisor_x}
        elongation = _normalize_scale(
            (scale_y / divisor_y, scale_x / divisor_x), image.ndim
        )
        larger_shape = resize_shape(
            shape, (divisor_y, divisor_x), divisor_x=divisor_x, divisor_y=divisor_y
        )
        # print('Downsampling: ', scale)
        # print('  divisor:', (divisor_y, divisor_x))
        # print('  elongation:', elongation)
        # print('  shape:', shape)
        # print('  larger_shape:', larger_shape)
        divisible_chunks = _make_divisible_tiles(larger_shape, divisor_x, divisor_y)
        image = _transform_array(
            image,
            elongation,
            offset,
            larger_shape,
            divisible_chunks,
            spline_order,
            recover_nan,
        )
        image = da.coarsen(aggregator, image, axes)
        if shape != image.shape:
            image = image[..., 0 : shape[-2], 0 : shape[-1]]
        if chunks is not None:
            image = image.rechunk(chunks)
    else:
        # Upsampling
        # ----------
        # print('Upsampling: ', scale)
        image = _transform_array(
            image, scale, offset, shape, chunks, spline_order, recover_nan
        )
    return image


def _transform_array(
    image: da.Array,
    scale: tuple[float, ...],
    offset: tuple[float, ...],
    shape: tuple[int, ...],
    chunks: Optional[tuple[int, ...]],
    spline_order: int,
    recover_nan: bool,
) -> da.Array:
    """Apply affine transformation to ND-image.

    Args:
        image: ND-image with shape (..., size_y, size_x)
        scale: Scaling factors (1, ..., 1, sy, sx)
        offset: Offset values (0, ..., 0, oy, ox)
        shape: (..., size_y, size_x)
        chunks: (..., chunk_size_y, chunk_size_x)
        spline_order: 0 ... 5
        recover_nan: True/False

    Returns:
        Transformed ND-image.
    """
    assert_true(len(scale) == image.ndim, "invalid scale")
    assert_true(len(offset) == image.ndim, "invalid offset")
    assert_true(len(shape) == image.ndim, "invalid shape")
    assert_true(chunks is None or len(chunks) == image.ndim, "invalid chunks")
    if _is_no_op(image, scale, offset, shape):
        return image
    # As of scipy 0.18, matrix = scale is no longer supported.
    # Therefore we use the diagonal matrix form here,
    # where scale is the diagonal.
    matrix = np.diag(scale)
    at_kwargs = dict(
        offset=offset,
        order=spline_order,
        output_shape=shape,
        output_chunks=chunks,
        mode="constant",
    )
    if recover_nan and spline_order > 0:
        # We can "recover" values that are neighbours to NaN values
        # that would otherwise become NaN too.
        mask = da.isnan(image)
        # First check if there are NaN values ar all
        if da.any(mask):
            # Yes, then
            # 1. replace NaN by zero
            filled_im = da.where(mask, 0.0, image)
            # 2. transform the zero-filled image
            scaled_im = ndinterp.affine_transform(
                filled_im, matrix, **at_kwargs, cval=0.0
            )
            # 3. transform the inverted mask
            scaled_norm = ndinterp.affine_transform(
                1.0 - mask, matrix, **at_kwargs, cval=0.0
            )
            # 4. put back NaN where there was zero,
            #    otherwise decode using scaled mask
            return da.where(
                da.isclose(scaled_norm, 0.0), np.nan, scaled_im / scaled_norm
            )

    # No dealing with NaN required
    return ndinterp.affine_transform(image, matrix, **at_kwargs, cval=np.nan)


def resize_shape(
    shape: Sequence[int],
    scale: Union[float, tuple[float, ...]],
    divisor_x: int = 1,
    divisor_y: int = 1,
) -> tuple[int, ...]:
    scale = _normalize_scale(scale, len(shape))
    height, width = shape[-2], shape[-1]
    scale_y, scale_x = scale[-2], scale[-1]
    wf = width * abs(scale_x)
    hf = height * abs(scale_y)
    w = divisor_x * math.ceil(wf / divisor_x)
    h = divisor_y * math.ceil(hf / divisor_y)
    return tuple(shape[0:-2]) + (h, w)


def _make_divisible_tiles(
    larger_shape: tuple[int, ...], divisor_x: int, divisor_y: int
) -> tuple[int, ...]:
    w = min(larger_shape[-1], divisor_x * ((2048 + divisor_x - 1) // divisor_x))
    h = min(larger_shape[-2], divisor_y * ((2048 + divisor_y - 1) // divisor_y))
    return (len(larger_shape) - 2) * (1,) + (h, w)


def _normalize_image(im: NDImage) -> da.Array:
    return da.asarray(im)


def _normalize_offset(
    offset: Optional[Sequence[float]], ndim: int
) -> tuple[float, ...]:
    return _normalize_pair(offset, 0.0, ndim, "offset")


def _normalize_scale(scale: Optional[Sequence[float]], ndim: int) -> tuple[float, ...]:
    return _normalize_pair(scale, 1.0, ndim, "scale")


def _normalize_pair(
    pair: Optional[Sequence[float]], default: float, ndim: int, name: str
) -> tuple[float, ...]:
    if pair is None:
        pair = [default, default]
    elif isinstance(pair, (int, float)):
        pair = [pair, pair]
    elif len(pair) != 2:
        raise ValueError(f"illegal image {name}")
    return (ndim - 2) * (default,) + tuple(pair)


def _normalize_shape(shape: Optional[Sequence[int]], im: NDImage) -> tuple[int, ...]:
    if shape is None:
        return im.shape
    if len(shape) != 2:
        raise ValueError("illegal image shape")
    return im.shape[0:-2] + tuple(shape)


def _normalize_chunks(
    chunks: Optional[Sequence[int]], shape: tuple[int, ...]
) -> Optional[tuple[int, ...]]:
    if chunks is None:
        return None
    if len(chunks) < 2 or len(chunks) > len(shape):
        raise ValueError("illegal image chunks")
    return (len(shape) - len(chunks)) * (1,) + tuple(chunks)


def _is_no_op(
    im: NDImage, scale: Sequence[float], offset: Sequence[float], shape: tuple[int, ...]
):
    return (
        shape == im.shape
        and all(math.isclose(s, 1) for s in scale)
        and all(math.isclose(o, 0) for o in offset)
    )



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

import itertools

import numpy as np
import pandas as pd
import pyproj
import xarray as xr


def new_cube(
    title="Test Cube",
    width=360,
    height=180,
    x_name="lon",
    y_name="lat",
    x_dtype="float64",
    y_dtype=None,
    x_units="degrees_east",
    y_units="degrees_north",
    x_res=1.0,
    y_res=None,
    x_start=-180.0,
    y_start=-90.0,
    inverse_y=False,
    time_name="time",
    time_dtype="datetime64[ns]",
    time_units="seconds since 1970-01-01T00:00:00",
    time_calendar="proleptic_gregorian",
    time_periods=5,
    time_freq="D",
    time_start="2010-01-01T00:00:00",
    use_cftime=False,
    drop_bounds=False,
    variables=None,
    crs=None,
    crs_name=None,
    time_encoding_dtype="int64",
):
    """Create a new empty cube. Useful for creating cubes templates with
    predefined coordinate variables and metadata. The function is also
    heavily used by xcube's unit tests.

    The values of the *variables* dictionary can be either constants,
    array-like objects, or functions that compute their return value from
    passed coordinate indexes. The expected signature is:::

        def my_func(time: int, y: int, x: int) -> Union[bool, int, float]

    Args:
        title: A title. Defaults to 'Test Cube'.
        width: Horizontal number of grid cells. Defaults to 360.
        height: Vertical number of grid cells. Defaults to 180.
        x_name: Name of the x coordinate variable. Defaults to 'lon'.
        y_name: Name of the y coordinate variable. Defaults to 'lat'.
        x_dtype: Data type of x coordinates. Defaults to 'float64'.
        y_dtype: Data type of y coordinates. Defaults to 'float64'.
        x_units: Units of the x coordinates. Defaults to 'degrees_east'.
        y_units: Units of the y coordinates. Defaults to
            'degrees_north'.
        x_start: Minimum x value. Defaults to -180.
        y_start: Minimum y value. Defaults to -90.
        x_res: Spatial resolution in x-direction. Defaults to 1.0.
        y_res: Spatial resolution in y-direction. Defaults to 1.0.
        inverse_y: Whether to create an inverse y axis. Defaults to
            False.
        time_name: Name of the time coordinate variable. Defaults to
            'time'.
        time_periods: Number of time steps. Defaults to 5.
        time_freq: Duration of each time step. Defaults to `1D'.
        time_start: First time value. Defaults to '2010-01-01T00:00:00'.
        time_dtype: Numpy data type for time coordinates. Defaults to
            'datetime64[s]'. If used, parameter 'use_cftime' must be
            False.
        time_units: Units for time coordinates. Defaults to 'seconds
            since 1970-01-01T00:00:00'.
        time_calendar: Calender for time coordinates. Defaults to
            'proleptic_gregorian'.
        use_cftime: If True, the time will be given as data types
            according to the 'cftime' package. If used, the
            time_calendar parameter must be also be given with an
            appropriate value such as 'gregorian' or 'julian'. If used,
            parameter 'time_dtype' must be None.
        drop_bounds: If True, coordinate bounds variables are not
            created. Defaults to False.
        variables: Dictionary of data variables to be added. None by
            default.
        crs: pyproj-compatible CRS string or instance of pyproj.CRS or
            None
        crs_name: Name of the variable that will hold the CRS
            information. Ignored, if *crs* is not given.
        time_encoding_dtype: data type used to encode the time variable
            when serializing the dataset

    Returns:
        A cube instance
    """
    y_dtype = y_dtype if y_dtype is not None else y_dtype
    y_res = y_res if y_res is not None else x_res
    if width < 0 or height < 0 or x_res <= 0.0 or y_res <= 0.0:
        raise ValueError()
    if time_periods < 0:
        raise ValueError()

    if use_cftime and time_dtype is not None:
        raise ValueError('If "use_cftime" is True, "time_dtype" must not be set.')

    x_is_lon = x_name == "lon" or x_units == "degrees_east"
    y_is_lat = y_name == "lat" or y_units == "degrees_north"

    x_end = x_start + width * x_res
    y_end = y_start + height * y_res

    x_res_05 = 0.5 * x_res
    y_res_05 = 0.5 * y_res

    x_data = np.linspace(x_start + x_res_05, x_end - x_res_05, width, dtype=x_dtype)
    y_data = np.linspace(y_start + y_res_05, y_end - y_res_05, height, dtype=y_dtype)

    x_var = xr.DataArray(x_data, dims=x_name, attrs=dict(units=x_units))
    y_var = xr.DataArray(y_data, dims=y_name, attrs=dict(units=y_units))
    if inverse_y:
        y_var = y_var[::-1]

    if x_is_lon:
        x_var.attrs.update(long_name="longitude", standard_name="longitude")
    else:
        x_var.attrs.update(
            long_name="x coordinate of projection",
            standard_name="projection_x_coordinate",
        )
    if y_is_lat:
        y_var.attrs.update(long_name="latitude", standard_name="latitude")
    else:
        y_var.attrs.update(
            long_name="y coordinate of projection",
            standard_name="projection_y_coordinate",
        )

    if use_cftime:
        time_data_p1 = xr.cftime_range(
            start=time_start,
            periods=time_periods + 1,
            freq=time_freq,
            calendar=time_calendar,
        ).values
    else:
        time_data_p1 = pd.date_range(
            start=time_start, periods=time_periods + 1, freq=time_freq
        ).values
        time_data_p1 = time_data_p1.astype(dtype=time_dtype)

    time_delta = time_data_p1[1] - time_data_p1[0]
    time_data = time_data_p1[0:-1] + time_delta // 2
    time_var = xr.DataArray(time_data, dims=time_name)
    time_var.encoding["units"] = time_units
    time_var.encoding["calendar"] = time_calendar
    time_var.encoding["dtype"] = time_encoding_dtype

    coords = {x_name: x_var, y_name: y_var, time_name: time_var}
    if not drop_bounds:
        x_bnds_name = f"{x_name}_bnds"
        y_bnds_name = f"{y_name}_bnds"
        time_bnds_name = f"{time_name}_bnds"

        bnds_dim = "bnds"

        x_bnds_data = np.zeros((width, 2), dtype=np.float64)
        x_bnds_data[:, 0] = np.linspace(x_start, x_end - x_res, width, dtype=x_dtype)
        x_bnds_data[:, 1] = np.linspace(x_start + x_res, x_end, width, dtype=x_dtype)
        y_bnds_data = np.zeros((height, 2), dtype=np.float64)
        y_bnds_data[:, 0] = np.linspace(y_start, y_end - x_res, height, dtype=y_dtype)
        y_bnds_data[:, 1] = np.linspace(y_start + x_res, y_end, height, dtype=y_dtype)
        if inverse_y:
            y_bnds_data = y_bnds_data[::-1, ::-1]

        x_bnds_var = xr.DataArray(
            x_bnds_data, dims=(x_name, bnds_dim), attrs=dict(units=x_units)
        )
        y_bnds_var = xr.DataArray(
            y_bnds_data, dims=(y_name, bnds_dim), attrs=dict(units=y_units)
        )

        x_var.attrs["bounds"] = x_bnds_name
        y_var.attrs["bounds"] = y_bnds_name

        time_bnds_data = np.zeros((time_periods, 2), dtype=time_data_p1.dtype)
        time_bnds_data[:, 0] = time_data_p1[:-1]
        time_bnds_data[:, 1] = time_data_p1[1:]
        time_bnds_var = xr.DataArray(time_bnds_data, dims=(time_name, bnds_dim))
        time_bnds_var.encoding["units"] = time_units
        time_bnds_var.encoding["calendar"] = time_calendar
        time_bnds_var.encoding["dtype"] = time_encoding_dtype

        time_var.attrs["bounds"] = time_bnds_name

        coords.update(
            {
                x_bnds_name: x_bnds_var,
                y_bnds_name: y_bnds_var,
                time_bnds_name: time_bnds_var,
            }
        )

    attrs = dict(
        Conventions="CF-1.7",
        title=title,
        time_coverage_start=str(time_data_p1[0]),
        time_coverage_end=str(time_data_p1[-1]),
    )

    if x_is_lon:
        attrs.update(
            dict(
                geospatial_lon_min=x_start,
                geospatial_lon_max=x_end,
                geospatial_lon_units=x_units,
            )
        )

    if y_is_lat:
        attrs.update(
            dict(
                geospatial_lat_min=y_start,
                geospatial_lat_max=y_end,
                geospatial_lat_units=y_units,
            )
        )

    data_vars = {}
    if variables:
        dims = (time_name, y_name, x_name)
        shape = (time_periods, height, width)
        size = time_periods * height * width
        for var_name, data in variables.items():
            if isinstance(data, xr.DataArray):
                data_vars[var_name] = data
            elif (
                isinstance(data, int)
                or isinstance(data, float)
                or isinstance(data, bool)
            ):
                data_vars[var_name] = xr.DataArray(np.full(shape, data), dims=dims)
            elif callable(data):
                func = data
                data = np.zeros(shape)
                for index in itertools.product(*map(range, shape)):
                    data[index] = func(*index)
                data_vars[var_name] = xr.DataArray(data, dims=dims)
            elif data is None:
                data_vars[var_name] = xr.DataArray(
                    np.random.uniform(0.0, 1.0, size).reshape(shape), dims=dims
                )
            else:
                data_vars[var_name] = xr.DataArray(data, dims=dims)

    if isinstance(crs, str):
        crs = pyproj.CRS.from_string(crs)

    if isinstance(crs, pyproj.CRS):
        crs_name = crs_name or "crs"
        for v in data_vars.values():
            v.attrs["grid_mapping"] = crs_name
        data_vars[crs_name] = xr.DataArray(0, attrs=crs.to_cf())

    return xr.Dataset(data_vars=data_vars, coords=coords, attrs=attrs)



#  Copyright (c) 2018-2025 by xcube team and contributors
#  Permissions are hereby granted under the terms of the MIT License:
#  https://opensource.org/licenses/MIT.

from chartlets import Component, Input, Output, State
from chartlets.components import Box, Checkbox, Select, Typography

from xcube.server.api import Context
from xcube.webapi.viewer.contrib import Panel, get_datasets_ctx

panel = Panel(__name__, title="Demo Panel", position=5)


COLORS = [(0, "red"), (1, "green"), (2, "blue"), (3, "yellow")]


@panel.layout(
    State("@app", "selectedDatasetId"),
)
def render_panel(
    ctx: Context,
    dataset_id: str = "",
) -> Component:
    opaque = False
    color = 0

    opaque_checkbox = Checkbox(
        id="opaque",
        value=opaque,
        label="Opaque",
    )

    color_select = Select(
        id="color",
        value=color,
        label="Color",
        options=COLORS,
        style={"flexGrow": 0, "minWidth": 80},
    )

    info_text = Typography(
        id="info_text", children=update_info_text(ctx, dataset_id, opaque, color)
    )

    return Box(
        style={
            "display": "flex",
            "flexDirection": "column",
            "width": "100%",
            "height": "100%",
            "gap": "6px",
        },
        children=[
            "This panel just demonstrates how server-side extensions work. "
            "It has no useful functionality.",
            opaque_checkbox,
            color_select,
            info_text
        ],
    )


# noinspection PyUnusedLocal
@panel.callback(
    Input("@app", "selectedDatasetId"),
    Input("opaque"),
    Input("color"),
    State("info_text", "children"),
    Output("info_text", "children"),
)
def update_info_text(
    ctx: Context,
    dataset_id: str = "",
    opaque: bool = False,
    color: int = 0,
    info_children: list[str] = "",
) -> list[str]:
    ds_ctx = get_datasets_ctx(ctx)
    ds_configs = ds_ctx.get_dataset_configs()

    info_text = info_children[0] if info_children else ""

    opaque = opaque or False
    color = color if color is not None else 0
    return [
        f"The dataset is {dataset_id},"
        f" the color is {COLORS[color][1]} and"
        f" it {'is' if opaque else 'is not'} opaque."
        f" The length of the last info text"
        f" was {len(info_text or '')}."
        f" The number of datasets is {len(ds_configs)}."
    ]



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

import logging
import sys
from collections.abc import Sequence
from typing import Any, Dict, Optional, Tuple, Type, Union

import click

from xcube.constants import (
    GENERAL_LOG_FORMAT,
    LOG,
    LOG_LEVEL_DETAIL,
    LOG_LEVEL_OFF,
    LOG_LEVEL_OFF_NAME,
    LOG_LEVEL_TRACE,
    XCUBE_LOG_FORMAT,
)


def new_cli_ctx_obj():
    return {
        "traceback": False,
        "scheduler": None,
    }


def cli_option_traceback(func):
    """Decorator for adding a reusable CLI option `--traceback`."""

    # noinspection PyUnusedLocal
    def _callback(ctx: click.Context, param: click.Option, value: bool):
        ctx_obj = ctx.ensure_object(dict)
        ctx_obj["traceback"] = value
        return value

    return click.option(
        "--traceback",
        is_flag=True,
        help="Enable tracing back errors by dumping the Python call stack. "
        "Pass as very first option to also trace back error during command-line validation.",
        callback=_callback,
    )(func)


def cli_option_quiet(func):
    """Decorator for adding a reusable CLI option `--quiet`/'-q'."""

    # noinspection PyUnusedLocal
    def _callback(ctx: click.Context, param: click.Option, value: bool):
        ctx_obj = ctx.ensure_object(dict)
        ctx_obj["quiet"] = value
        return value

    return click.option(
        "--quiet",
        "-q",
        is_flag=True,
        help="Disable output of log messages to the console entirely."
        " Note, this will also suppress error and warning messages.",
        callback=_callback,
    )(func)


def cli_option_verbosity(func):
    """Decorator for adding a reusable CLI option `--verbose`/'-v' that
    can be used multiple times.
    The related kwarg is named `verbosity` and is of type int (= count).
    """

    # noinspection PyUnusedLocal
    def _callback(ctx: click.Context, param: click.Option, value: int):
        ctx_obj = ctx.ensure_object(dict)
        ctx_obj["verbose"] = value
        ctx_obj["verbosity"] = value
        return value

    return click.option(
        "--verbose",
        "-v",
        "verbosity",
        count=True,
        help="Enable output of log messages to the console."
        " Has no effect if --quiet/-q is used."
        " May be given multiple times to control the level"
        " of log messages, i.e.,"
        " -v refers to level INFO, -vv to DETAIL, -vvv to DEBUG,"
        " -vvvv to TRACE."
        " If omitted, the log level of the console is WARNING.",
        callback=_callback,
    )(func)


def cli_option_dry_run(func):
    """Decorator for adding a reusable CLI option `--dry-run`/'-d'.
    The related kwarg is named `dry_run` and is of type bool.
    """

    # noinspection PyUnusedLocal
    def _callback(ctx: click.Context, param: click.Option, value: bool):
        ctx_obj = ctx.ensure_object(dict)
        ctx_obj["dry_run"] = value
        return value

    return click.option(
        "--dry-run",
        "-d",
        "dry_run",
        is_flag=True,
        help="Do not change any data, just report what would have been changed.",
        callback=_callback,
    )(func)


def cli_option_scheduler(func):
    """Decorator for adding a pre-defined, reusable CLI option `--scheduler`."""

    # noinspection PyUnusedLocal
    def _callback(ctx: click.Context, param: click.Option, value: Optional[str]):
        if not value:
            return

        address_and_kwargs = value.split("?", 2)
        if len(address_and_kwargs) == 2:
            address, kwargs_string = address_and_kwargs
            kwargs = parse_cli_kwargs(kwargs_string, metavar="SCHEDULER")
        else:
            (address,) = address_and_kwargs
            kwargs = dict()

        try:
            # The Dask Client registers itself as the default Dask scheduler, and so runs dask.array used by xarray
            import distributed

            scheduler_client = distributed.Client(address, **kwargs)
            ctx_obj = ctx.ensure_object(dict)
            ctx_obj["scheduler"] = scheduler_client
            return scheduler_client
        except ValueError as e:
            raise click.BadParameter(
                f"Failed to create Dask scheduler client: {e}"
            ) from e

    return click.option(
        "--scheduler",
        metavar="SCHEDULER",
        help="Enable distributed computing using the Dask scheduler identified by SCHEDULER. "
        "SCHEDULER can have the form <address>?<keyword>=<value>,... where <address> "
        "is <host> or <host>:<port> and specifies the scheduler's address in your network. "
        "For more information on distributed computing "
        "using Dask, refer to http://distributed.dask.org/. "
        "Pairs of <keyword>=<value> are passed to the Dask client. "
        "Refer to http://distributed.dask.org/en/latest/api.html#distributed.Client",
        callback=_callback,
    )(func)


def parse_cli_kwargs(value: str, metavar: str = None) -> dict[str, Any]:
    """Parse a string value of the form [<kw>=<arg>{,<kw>=<arg>}] into a dictionary.
    <kw> must be a valid Python identifier, <arg> must be a Python literal.

    Args:
        value: A string value.
        metavar: Name of a meta-variable used in CLI.

    Returns:
        a dictionary of keyword-arguments
    """
    if value:
        try:
            return eval(f"dict({value})", {}, {})
        except Exception:
            if metavar:
                message = f"Invalid value for {metavar}: {value!r}"
            else:
                message = f"Invalid value: {value!r}"
            raise click.ClickException(message)
    else:
        return dict()


def parse_cli_sequence(
    seq_value: Union[None, str, Sequence[Any]],
    metavar: str = "parameter",
    item_parser=None,
    item_validator=None,
    allow_none: bool = True,
    allow_empty_items: bool = False,
    strip_items: bool = True,
    item_plural_name: str = "items",
    num_items: int = None,
    num_items_min: int = None,
    num_items_max: int = None,
    separator: str = ",",
    error_type: type[Exception] = click.ClickException,
) -> Optional[tuple[Any, ...]]:
    """Parse a CLI argument that is supposed to be a sequence.

    Args:
        seq_value: A string, a string sequence, or None.
        metavar: CLI metavar name
        item_parser: an optional function that takes a string and parses
            it
        item_validator: a optional function that takes a parsed value
            and parses it
        allow_none: whether it is ok for *seq_value* to be None
        allow_empty_items: whether it is ok to have empty values in
            *seq_value*
        strip_items: whether to strip values in *seq_value*
        item_plural_name: a name for multiple items
        num_items: expected number of items
        num_items_min: expected minimum number of items
        num_items_max: expected maximum number of items
        separator: expected separator if *seq_value* is a string,
            default is ','.
        error_type: value error to be raised in case, defaults to
            ``click.ClickException``

    Returns:
        parsed and validated *seq_value* as a tuple of values
    """
    if seq_value is None:
        if allow_none:
            return None
        raise error_type(f"{metavar} must be given")
    if isinstance(seq_value, str):
        if "," in seq_value:
            items = seq_value.split(",")
        elif num_items is not None:
            items = num_items * (seq_value,)
        else:
            items = (seq_value,)
    else:
        items = seq_value
    item_count = len(items)
    if num_items is not None and item_count != num_items:
        raise error_type(
            f"{metavar} must have {num_items} {item_plural_name} separated by {separator!r}"
        )
    if num_items_min is not None and item_count < num_items_min:
        raise error_type(
            f"{metavar} must have at least {num_items_min} {item_plural_name} separated by {separator!r}"
        )
    if num_items_max is not None and item_count > num_items_max:
        raise error_type(
            f"{metavar} must have no more than {num_items_max} {item_plural_name} separated by {separator!r}"
        )
    if strip_items:
        items = tuple(item.strip() for item in items)
    if not allow_empty_items:
        for item in items:
            if not item:
                raise error_type(f"{item_plural_name} in {metavar} must not be empty")
    if item_parser:
        try:
            items = tuple(map(item_parser, items))
        except ValueError as e:
            raise error_type(f"Invalid {item_plural_name} in {metavar} found: {e}")
    if item_validator:
        try:
            for item in items:
                item_validator(item)
        except ValueError as e:
            raise error_type(f"Invalid {item_plural_name} in {metavar} found: {e}")
    return tuple(items)


def assert_positive_int_item(item: int):
    """A validator for positive integer number sequences.
    Frequently used to validate counts or image and tile sizes passes as args to the CLI.
    """
    if item <= 0:
        raise ValueError("all items must be positive integer numbers")


def handle_cli_exception(
    e: BaseException, exit_code: int = None, traceback_mode: bool = False
) -> int:
    exc_info = traceback_mode and e
    if isinstance(e, click.Abort):
        LOG.error("Aborted.", exc_info=exc_info)
        exit_code = exit_code or 1
    elif isinstance(e, click.ClickException):
        LOG.error("%s", e, exc_info=exc_info)
        exit_code = exit_code or e.exit_code
    elif isinstance(e, OSError):
        LOG.error("OS error: %s", e, exc_info=exc_info)
        exit_code = exit_code or 2
    else:
        LOG.error(f"Internal error: %s", e, exc_info=exc_info)
        exit_code = exit_code or 3
    LOG.debug("Exit with code %d", exit_code)
    return exit_code


def configure_warnings(enable_warnings: bool):
    # In normal operation, it's not necessary to explicitly set the default
    # filter when --warnings is omitted, but it can be needed during
    # unit testing if a previous test has caused the filter to be changed.
    import warnings

    warnings.simplefilter(
        "default" if enable_warnings else "ignore", category=DeprecationWarning
    )
    warnings.simplefilter(
        "default" if enable_warnings else "ignore", category=RuntimeWarning
    )


_general_handler: Union[
    logging.NullHandler, logging.FileHandler, logging.StreamHandler
] = logging.NullHandler()


def configure_logging(
    log_file: Optional[str],
    log_level: Optional[str],
    logger: logging.Logger = logging.getLogger(),
):
    remove_log_handlers(logger)
    if log_level == LOG_LEVEL_OFF_NAME:
        logger.setLevel(LOG_LEVEL_OFF)
    else:
        logger.setLevel(log_level)
        formatter = logging.Formatter(GENERAL_LOG_FORMAT)
        if log_file:
            handler = logging.FileHandler(log_file, "a", encoding="utf8")
        else:
            handler = logging.StreamHandler(sys.stderr)
        handler.setFormatter(formatter)
        logger.addHandler(handler)
        global _general_handler
        _general_handler = handler


def configure_cli_output(
    quiet: Optional[bool] = None,
    verbosity: Optional[Union[bool, int]] = None,
    logger: logging.Logger = LOG,
):
    remove_log_handlers(logger)

    if quiet:
        level = LOG_LEVEL_OFF
    elif not verbosity:
        level = logging.WARNING
    elif verbosity == 1:
        level = logging.INFO
    elif verbosity == 2:
        level = LOG_LEVEL_DETAIL
    elif verbosity == 3:
        level = logging.DEBUG
    else:
        level = LOG_LEVEL_TRACE

    logger.setLevel(level)

    if isinstance(_general_handler, (logging.FileHandler, logging.NullHandler)):
        # Only if we do not already redirect output of general logging
        # to stderr, install a new handler with a simple message format
        # for the console.
        handler = logging.StreamHandler(sys.stderr)
        handler.setFormatter(ConsoleMessageFormatter())
        logger.addHandler(handler)


def remove_log_handlers(logger: logging.Logger):
    for h in list(logger.handlers):
        logger.removeHandler(h)
        h.close()


class ConsoleMessageFormatter(logging.Formatter):
    def __init__(self):
        super().__init__(XCUBE_LOG_FORMAT)

    def format(self, record: logging.LogRecord) -> str:
        if logging.DEBUG < record.levelno < logging.WARNING:
            # Just return plain message and skip formatting level
            msg = str(record.msg)
            return msg % record.args if record.args else msg
        else:
            return super().format(record)



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

import concurrent.futures
from collections.abc import Awaitable, Sequence
from typing import Any, Callable, Optional, Tuple, Union

from xcube.server.api import ApiRoute, ApiStaticRoute, Context, ReturnT
from xcube.server.framework import Framework
from xcube.util.jsonschema import JsonObjectSchema


class FlaskFramework(Framework):
    """The Flask web server framework.

    TODO: implement me!
    """

    @property
    def config_schema(self) -> Optional[JsonObjectSchema]:
        return None

    def add_static_routes(
        self, static_routes: Sequence[ApiStaticRoute], url_prefix: str
    ):
        raise NotImplementedError()

    def add_routes(self, routes: Sequence[ApiRoute], url_prefix: str):
        raise NotImplementedError()

    def update(self, ctx: Context):
        raise NotImplementedError()

    def start(self, ctx: Context):
        raise NotImplementedError()

    def stop(self, ctx: Context):
        raise NotImplementedError()

    def call_later(
        self, delay: Union[int, float], callback: Callable, *args, **kwargs
    ) -> object:
        raise NotImplementedError()

    def run_in_executor(
        self,
        executor: Optional[concurrent.futures.Executor],
        function: Callable[..., ReturnT],
        *args: Any,
        **kwargs: Any,
    ) -> Awaitable[ReturnT]:
        raise NotImplementedError()



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

from .abc import MultiLevelDataset
from .mapped import MappedMultiLevelDataset


class IdentityMultiLevelDataset(MappedMultiLevelDataset):
    """The identity."""

    def __init__(self, ml_dataset: MultiLevelDataset, ds_id: str = None):
        super().__init__(ml_dataset, lambda ds: ds, ds_id=ds_id)



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

from xcube.server.api import ApiError, ApiHandler

from .api import api
from .context import PlacesContext
from .controllers import find_places

PATH_PARAM_PLACE_GROUP_ID = {
    "name": "placeGroupId",
    "in": "path",
    "description": "Place group identifier",
    "schema": {"type": "string"},
}


@api.route("/places")
class PlaceGroupsHandler(ApiHandler[PlacesContext]):
    @api.operation(
        operationId="getPlaceGroups", summary="Get place groups including all places."
    )
    def get(self):
        place_groups = self.ctx.get_global_place_groups(self.request.reverse_base_url)
        self.response.finish({"placeGroups": place_groups})


# noinspection PyPep8Naming
@api.route("/places/{placeGroupId}")
class FindPlacesHandler(ApiHandler):
    """Find places within a known place group."""

    @api.operation(
        operationId="findPlacesInPlaceGroup",
        summary="Find places in a given place group.",
        parameters=[
            PATH_PARAM_PLACE_GROUP_ID,
            {
                "name": "geom",
                "in": "query",
                "description": "Geometry WKT string",
                "schema": {"type": "string"},
            },
            {
                "name": "bbox",
                "in": "query",
                "description": "Bounding box string in the form x1,y1,x2,y2",
                "schema": {"type": "string"},
            },
        ],
    )
    def get(self, placeGroupId: str):
        geom_wkt = self.request.get_query_arg("geom", default=None)
        box_coords = self.request.get_query_arg("bbox", default=None)
        # Not implemented yet:
        # query_expr = self.request.get_query_arg("query", default=None)
        # comb_op = self.request.get_query_arg("comb", default="and")
        if geom_wkt and box_coords:
            raise ApiError.BadRequest('Only one of "geom" and "bbox" may be given')
        places = find_places(
            self.ctx,
            placeGroupId,
            self.request.reverse_base_url,
            query_geometry=box_coords or geom_wkt or None,
            # query_expr=query_expr,
            # comb_op=comb_op
        )
        self.response.finish({"places": places})

    @api.operation(
        operationId="findPlacesInPlaceGroup",
        summary="Find places in a given place group for a GeoJSON object.",
        parameters=[PATH_PARAM_PLACE_GROUP_ID],
        description="The request body must be a GeoJSON object.",
    )
    def post(self, placeGroupId: str):
        # Not implemented yet:
        # query_expr = self.request.get_query_arg("query", default=None)
        # comb_op = self.request.get_query_arg("comb", default="and")
        geojson_obj = self.request.json
        places = find_places(
            self.ctx,
            placeGroupId,
            self.request.reverse_base_url,
            query_geometry=geojson_obj,
            # query_expr=query_expr,
            # comb_op=comb_op
        )
        self.response.finish({"places": places})



# Copyright (c) 2018-2025 by xcube team and contributors
# Permissions are hereby granted under the terms of the MIT License:
# https://opensource.org/licenses/MIT.

import warnings
from collections.abc import Sequence
from typing import Any, Dict

import click

from xcube.cli.common import (
    cli_option_quiet,
    cli_option_verbosity,
    configure_cli_output,
)
from xcube.constants import FORMAT_NAME_MEM, FORMAT_NAME_NETCDF4, FORMAT_NAME_ZARR, LOG

UPSAMPLING_METHODS = ["asfreq", "ffill", "bfill", "pad", "nearest", "interpolate"]
DOWNSAMPLING_METHODS = [
    "count",
    "first",
    "last",
    "min",
    "max",
    "sum",
    "prod",
    "mean",
    "median",
    "std",
    "var",
]
RESAMPLING_METHODS = UPSAMPLING_METHODS + DOWNSAMPLING_METHODS

SPLINE_INTERPOLATION_KINDS = ["zero", "slinear", "quadratic", "cubic"]
OTHER_INTERPOLATION_KINDS = ["linear", "nearest", "previous", "next"]
INTERPOLATION_KINDS = SPLINE_INTERPOLATION_KINDS + OTHER_INTERPOLATION_KINDS

OUTPUT_FORMAT_NAMES = [FORMAT_NAME_ZARR, FORMAT_NAME_NETCDF4, FORMAT_NAME_MEM]

DEFAULT_OUTPUT_PATH = "out.zarr"
DEFAULT_RESAMPLING_METHOD = "mean"
DEFAULT_RESAMPLING_FREQUENCY = "1D"
DEFAULT_RESAMPLING_BASE = None  # Deprecated since 1.0.4!
DEFAULT_INTERPOLATION_KIND = "linear"


# noinspection PyShadowingBuiltins
@click.command(name="resample")
@click.argument("cube")
@click.option(
    "--config",
    "-c",
    metavar="CONFIG",
    multiple=True,
    help="xcube dataset configuration file in YAML format. More than one config input file is allowed."
    "When passing several config files, they are merged considering the order passed via command line.",
)
@click.option(
    "--output",
    "-o",
    metavar="OUTPUT",
    default=DEFAULT_OUTPUT_PATH,
    help=f"Output path. Defaults to {DEFAULT_OUTPUT_PATH!r}.",
)
@click.option(
    "--format",
    "-f",
    type=click.Choice(OUTPUT_FORMAT_NAMES),
    help="Output format. If omitted, format will be guessed from output path.",
)
@click.option(
    "--variables",
    "--vars",
    metavar="VARIABLES",
    help="Comma-separated list of names of variables to be included.",
)
@click.option(
    "--method",
    "-M",
    multiple=True,
    help=f"Temporal resampling method. "
    f"Available downsampling methods are "
    f"{', '.join(map(repr, DOWNSAMPLING_METHODS))}, "
    f"the upsampling methods are "
    f"{', '.join(map(repr, UPSAMPLING_METHODS))}. "
    f"If the upsampling method is 'interpolate', "
    f"the option '--kind' will be used, if given. "
    f"Other upsampling methods that select existing values "
    f"honour the '--tolerance' option. "
    f"Defaults to {DEFAULT_RESAMPLING_METHOD!r}.",
)
@click.option(
    "--frequency",
    "-F",
    help='Temporal aggregation frequency. Use format "<count><offset>" '
    "where <offset> is one of 'H', 'D', 'W', 'M', 'Q', 'Y'. "
    "Use 'all' to aggregate all time steps included in the dataset."
    f"Defaults to {DEFAULT_RESAMPLING_FREQUENCY!r}.",
)
@click.option(
    "--offset",
    "-O",
    help="Offset used to adjust the resampled time labels. Uses same syntax as frequency. "
    "Some Pandas date offset strings are supported as well.",
)
@click.option(
    "--kind",
    "-K",
    type=str,
    default=DEFAULT_INTERPOLATION_KIND,
    help="Interpolation kind which will be used if upsampling method is 'interpolation'. "
    f"May be one of {', '.join(map(repr, INTERPOLATION_KINDS))} where "
    f"{', '.join(map(repr, SPLINE_INTERPOLATION_KINDS))} refer to a spline interpolation of "
    f"zeroth, first, second or third order; 'previous' and 'next' "
    f"simply return the previous or next value of the point. "
    "For more info "
    "refer to scipy.interpolate.interp1d(). "
    f"Defaults to {DEFAULT_INTERPOLATION_KIND!r}.",
)
@click.option(
    "--tolerance",
    "-T",
    type=str,
    help="Tolerance for selective upsampling methods. Uses same syntax as frequency. "
    "If the time delta exceeds the tolerance, "
    "fill values (NaN) will be used. "
    "Defaults to the given frequency.",
)
@cli_option_quiet
@cli_option_verbosity
@click.option(
    "--dry-run",
    default=False,
    is_flag=True,
    help="Just read and process inputs, but don't produce any outputs.",
)
def resample(
    cube,
    config,
    output,
    format,
    variables,
    method,
    frequency,
    offset,
    kind,
    tolerance,
    quiet,
    verbosity,
    dry_run,
):
    """Resample data along the time dimension."""
    configure_cli_output(quiet=quiet, verbosity=verbosity)

    input_path = cube
    config_files = config
    output_path = output
    output_format = format

    from xcube.util.config import load_configs

    config = load_configs(*config_files) if config_files else {}

    if input_path:
        config["input_path"] = input_path
    if output_path:
        config["output_path"] = output_path
    if output_format:
        config["output_format"] = output_format
    if method:
        config["methods"] = method
    if frequency:
        config["frequency"] = frequency
    if offset:
        config["offset"] = offset
    if kind:
        config["interp_kind"] = kind
    if tolerance:
        config["tolerance"] = tolerance
    if variables:
        try:
            variables = set(map(lambda c: str(c).strip(), variables.split(",")))
        except ValueError:
            variables = None
        if variables is not None and next(
            iter(True for var_name in variables if var_name == ""), False
        ):
            variables = None
        if variables is None or len(variables) == 0:
            raise click.ClickException(f"invalid variables {variables!r}")
        config["variables"] = variables

    if "methods" in config:
        methods = config["methods"]
        for method in methods:
            if method not in RESAMPLING_METHODS:
                raise click.ClickException(f"invalid resampling method {method!r}")

    # noinspection PyBroadException
    _resample_in_time(**config, dry_run=dry_run, monitor=LOG.info)

    return 0


def _resample_in_time(
    input_path: str = None,
    variables: Sequence[str] = None,
    metadata: dict[str, Any] = None,
    output_path: str = DEFAULT_OUTPUT_PATH,
    output_format: str = None,
    methods: Sequence[str] = (DEFAULT_RESAMPLING_METHOD,),
    frequency: str = DEFAULT_RESAMPLING_FREQUENCY,
    offset: str = None,
    interp_kind: str = DEFAULT_INTERPOLATION_KIND,
    tolerance: str = None,
    dry_run: bool = False,
    monitor=None,
):
    from xcube.core.dsio import guess_dataset_format, open_cube, write_cube
    from xcube.core.resampling import resample_in_time
    from xcube.core.update import update_dataset_chunk_encoding

    if not output_format:
        output_format = guess_dataset_format(output_path)

    monitor(f"Opening cube from {input_path!r}...")
    with open_cube(input_path) as ds:
        monitor("Resampling...")
        agg_ds = resample_in_time(
            ds,
            frequency=frequency,
            method=methods,
            offset=offset,
            interp_kind=interp_kind,
            tolerance=tolerance,
            time_chunk_size=1,
            var_names=variables,
            metadata=metadata,
        )

        agg_ds = update_dataset_chunk_encoding(
            agg_ds, chunk_sizes={}, format_name=output_format, in_place=True
        )

        monitor(f"Writing resampled cube to {output_path!r}...")
        if not dry_run:
            write_cube(agg_ds, output_path, output_format, cube_asserted=True)

        monitor(f"Done.")
